from __future__ import annotations
from typing import Any, Iterable, MutableMapping
from qtpy.QtWidgets import QTreeWidget, QTreeWidgetItem

from .object import BaseWidget, ContextMenuMixin, PyObjectBound

# WIP!


class DictTreeWidget(BaseWidget, MutableMapping):
    def __init__(self, value=None, **kwargs):
        super().__init__(**kwargs)

        self._treewidget = PyTreeWidget(self.native)
        self._treewidget.setParentWidget(self)
        self._dict: dict[str, int] = {}  # mapping from key to row

        self.set_widget(self._treewidget)

        @self._treewidget.itemDoubleClicked.connect
        def _(item: PyTreeWidgetItem):
            type_ = type(item.obj)
            callbacks = self._callbacks.get(type_, [])
            self.running = True
            try:
                for callback in callbacks:
                    try:
                        callback(item.obj, item)
                    except TypeError:
                        callback(item.obj)
            finally:
                self.running = False

        if value is not None:
            self.update(dict(value))

    def __len__(self) -> int:
        return self._treewidget.rowCount()

    @property
    def value(self) -> dict[str, Any]:
        return {k: self._treewidget.item(row, 0) for k, row in self._dict}

    def __getitem__(self, k: str) -> Any:
        row = self._dict[k]
        return self._treewidget.item(row, 0).obj

    def __setitem__(self, k: str, obj: Any) -> None:
        if not isinstance(k, str):
            raise ValueError("Can only use str type as keys.")

        if k in self._dict.keys():
            row = self._dict[k]
        else:
            row = len(self)
            self._dict[k] = row
            key_item = QTreeWidgetItem(k)

        name = self._delegates.get(type(obj), str)(obj)
        value_item = PyTreeWidgetItem(obj, name)
        tooltip = self._tooltip.get(type(obj), str)(obj)
        value_item.setToolTip(tooltip)

    def __delitem__(self, k: str) -> None:
        row = self._dict.pop(k)

    def __iter__(self) -> Iterable[str]:
        return iter(self._dict)

    def keys(self):
        """
        Return the view of dictionary keys.
        """
        return self._dict.keys()

    def values(self) -> DictValueView:
        """
        Return the view of dictionary values as Python objects.
        """
        return DictValueView(self._treewidget)

    # def items(self) -> DictItemView:
    #     """
    #     Return the view of dictionary keys and values as strings and Python objects.
    #     """
    #     return DictItemView(self._treewidget)

    def update(self, d: dict[str, Any]):
        """
        Update the dictionary contents.
        """
        for k, v in d.items():
            self[k] = v

    def clear(self) -> None:
        """
        Clear dictionary contents.
        """
        self._treewidget.clear()
        self._dict.clear()

    def pop(self, k: str):
        """
        Pop a dictionary content.
        """
        row = self._dict.pop(k)
        out = self._treewidget.item(row, 0).obj
        self._treewidget.removeRow(row)
        return out

    def get(self, k: str, default=None):
        self._dict.get(k, default)


class PyTreeWidget(ContextMenuMixin, QTreeWidget):
    def item(self, row: int, column: int) -> PyTreeWidgetItem:
        return super().item(row, column)

    def itemAt(self, *p) -> PyTreeWidgetItem:
        return super().itemAt(*p)

    def __init__(self, parent: None) -> None:
        super().__init__(parent=parent)
        self.setContextMenu()


unbound = object()


class PyTreeWidgetItem(PyObjectBound, QTreeWidgetItem):
    def __init__(self, obj=None, name=None):
        super().__init__()
        self.setObject(obj, name)


class DictValueView:
    def __init__(self, widget: PyTreeWidget):
        self.widget = widget

    def __iter__(self):
        for row in range(self.widget.rowCount()):
            yield self.widget.item(row, 0).obj


# class DictItemView:
#     def __init__(self, widget: PyTableWidget):
#         self.widget = widget

#     def __iter__(self):
#         for row in range(self.widget.rowCount()):
#             key = self.widget.verticalHeaderItem(row).text()
#             value = self.widget.item(row, 0).obj
#             yield key, value



from __future__ import annotations
import macrokit as mk
from macrokit import Expr, Head, symbol
from enum import Enum
from pathlib import Path
import datetime
from .types import Color
from .widgets import ColorEdit
from .widgets.sequence import ListDataView
from ._gui._base import MagicTemplate

# classes
_datetime = Expr(Head.getattr, [datetime, datetime.datetime])
_date = Expr(Head.getattr, [datetime, datetime.date])
_time = Expr(Head.getattr, [datetime, datetime.time])

# magicgui-style input
mk.register_type(Enum, lambda e: symbol(e.value))
mk.register_type(Path, lambda e: f"r'{e}'")
mk.register_type(ListDataView, lambda e: list(e))
mk.register_type(
    datetime.datetime,
    lambda e: Expr.parse_call(
        _datetime, (e.year, e.month, e.day, e.hour, e.minute), {}
    ),
)
mk.register_type(
    datetime.date, lambda e: Expr.parse_call(_date, (e.year, e.month, e.day), {})
)
mk.register_type(
    datetime.time, lambda e: Expr.parse_call(_time, (e.hour, e.minute), {})
)


@mk.register_type(MagicTemplate)
def find_myname(gui: MagicTemplate):
    """This function is the essential part of macro recording"""
    parent = gui.__magicclass_parent__
    if parent is None:
        return gui._my_symbol
    else:
        return Expr(Head.getattr, [find_myname(parent), gui._my_symbol])


import magicgui as mgui

if tuple(int(v) for v in mgui.__version__.split(".")[:2]) < (0, 4):
    # magicgui<0.4 has bug in type registration
    from magicgui.type_map import _TYPE_DEFS

    _TYPE_DEFS[Color] = (ColorEdit, {})
else:
    mgui.register_type(Color, widget_type=ColorEdit)



from __future__ import annotations
from typing import Callable, TYPE_CHECKING, Any
import warnings
from magicgui.widgets import Image, Table, Label, FunctionGui
from magicgui.widgets._bases import ButtonWidget
from magicgui.widgets._bases.widget import Widget
from macrokit import Symbol
from qtpy.QtWidgets import QToolBar, QMenu, QWidgetAction, QTabWidget


from .mgui_ext import AbstractAction, _LabeledWidgetAction, WidgetAction, ToolButtonPlus
from .keybinding import register_shortcut
from ._base import (
    BaseGui,
    PopUpMode,
    ErrorMode,
    ContainerLikeGui,
    nested_function_gui_callback,
    _inject_recorder,
)
from .utils import copy_class, format_error, set_context_menu
from .menu_gui import ContextMenuGui, MenuGui, MenuGuiBase, insert_action_like

from ..signature import get_additional_option
from ..fields import MagicField
from ..widgets import FreeWidget, Separator
from ..utils import iter_members

if TYPE_CHECKING:
    from napari.viewer import Viewer


def _check_popupmode(popup_mode: PopUpMode):
    if popup_mode in (PopUpMode.above, PopUpMode.below, PopUpMode.first):
        msg = (
            f"magictoolbar does not support popup mode {popup_mode.value}."
            "PopUpMode.popup is used instead"
        )
        warnings.warn(msg, UserWarning)
    elif popup_mode == PopUpMode.last:
        msg = (
            f"magictoolbat does not support popup mode {popup_mode.value}."
            "PopUpMode.parentlast is used instead"
        )
        warnings.warn(msg, UserWarning)
        popup_mode = PopUpMode.parentlast

    return popup_mode


class QtTabToolBar(QToolBar):
    """ToolBar widget with tabs."""

    def __init__(self, title: str, parent=None):
        super().__init__(title, parent)
        self._tab = QTabWidget(self)
        self._tab.setContentsMargins(0, 0, 0, 0)
        self._tab.setTabBarAutoHide(True)
        self._tab.setStyleSheet(
            "QTabWidget {" "    margin: 0px, 0px, 0px, 0px;" "    padding: 0px;}"
        )
        self.addWidget(self._tab)

    def addToolBar(self, toolbar: QToolBar, name: str) -> None:
        """Add a toolbar as a tab."""
        self._tab.addTab(toolbar, name)
        toolbar.setContentsMargins(0, 0, 0, 0)
        return None


class ToolBarGui(ContainerLikeGui):
    """Magic class that will be converted into a toolbar"""

    def __init__(
        self,
        parent=None,
        name: str = None,
        close_on_run: bool = None,
        popup_mode: str | PopUpMode = None,
        error_mode: str | ErrorMode = None,
        labels: bool = True,
    ):
        popup_mode = _check_popupmode(popup_mode)

        super().__init__(
            close_on_run=close_on_run, popup_mode=popup_mode, error_mode=error_mode
        )
        name = name or self.__class__.__name__
        self.native = QToolBar(name, parent)
        self.name = name
        self._list: list[MenuGuiBase | AbstractAction] = []
        self.labels = labels

    def reset_choices(self, *_: Any):
        """Reset child Categorical widgets"""
        super().reset_choices()

        # If parent magic-class is added to napari viewer, the style sheet need update because
        # QToolButton has inappropriate style.
        # Detecting this event using "reset_choices" is not a elegant way, but works for now.
        viewer = self.parent_viewer
        if viewer is not None:
            style = _create_stylesheet(viewer)
            self.native.setStyleSheet(style)
        return None

    def _convert_attributes_into_widgets(self):
        cls = self.__class__

        # Add class docstring as label.
        if cls.__doc__:
            self.native.setToolTip(cls.__doc__)

        # Bind all the methods and annotations
        base_members = {x[0] for x in iter_members(ToolBarGui)}

        _hist: list[tuple[str, str, str]] = []  # for traceback
        _ignore_types = (property, classmethod, staticmethod)

        for name, attr in filter(lambda x: x[0] not in base_members, iter_members(cls)):
            if isinstance(attr, _ignore_types):
                continue

            try:
                if isinstance(attr, type):
                    # Nested magic-menu
                    widget = attr()
                    object.__setattr__(self, name, widget)

                elif isinstance(attr, MagicField):
                    widget = self._create_widget_from_field(name, attr)

                else:
                    # convert class method into instance method
                    widget = getattr(self, name, None)

                if isinstance(widget, FunctionGui):
                    widget[0].bind(self)  # set self to the first argument

                elif isinstance(widget, BaseGui):
                    widget.__magicclass_parent__ = self
                    self.__magicclass_children__.append(widget)
                    widget._my_symbol = Symbol(name)

                    if isinstance(widget, MenuGui):
                        tb = ToolButtonPlus(widget.name)
                        tb.set_menu(widget.native)
                        widget.native.setParent(tb.native, widget.native.windowFlags())
                        tb.tooltip = widget.__doc__
                        widget = WidgetAction(tb)

                    elif isinstance(widget, ContextMenuGui):
                        # Add context menu to toolbar
                        set_context_menu(widget, self)
                        _hist.append((name, type(attr), "ContextMenuGui"))

                    elif isinstance(widget, ToolBarGui):
                        tb = ToolButtonPlus(widget.name)
                        tb.tooltip = widget.__doc__
                        qmenu = QMenu(self.native)
                        waction = QWidgetAction(qmenu)
                        waction.setDefaultWidget(widget.native)
                        qmenu.addAction(waction)
                        tb.set_menu(qmenu)
                        widget = WidgetAction(tb)

                    else:
                        widget = WidgetAction(widget)

                elif isinstance(widget, Widget):
                    widget = WidgetAction(widget)

                if isinstance(widget, (AbstractAction, Callable, Widget)):
                    if (not isinstance(widget, Widget)) and callable(widget):
                        if name.startswith("_") or not get_additional_option(
                            attr, "gui", True
                        ):
                            keybinding = get_additional_option(attr, "keybinding", None)
                            if keybinding:
                                register_shortcut(
                                    keys=keybinding, parent=self.native, target=widget
                                )
                            continue
                        widget = self._create_widget_from_method(widget)

                    elif hasattr(widget, "__magicclass_parent__") or hasattr(
                        widget.__class__, "__magicclass_parent__"
                    ):
                        if isinstance(widget, BaseGui):
                            widget._my_symbol = Symbol(name)
                        # magic-class has to know its parent.
                        # if __magicclass_parent__ is defined as a property, hasattr must be called
                        # with a type object (not instance).
                        widget.__magicclass_parent__ = self

                    if name.startswith("_"):
                        continue

                    moveto = get_additional_option(attr, "into")
                    copyto = get_additional_option(attr, "copyto", [])
                    if moveto is not None or copyto:
                        self._unwrap_method(name, widget, moveto, copyto)
                    else:
                        self.insert(len(self), widget)

                    _hist.append((name, str(type(attr)), type(widget).__name__))

            except Exception as e:
                format_error(e, _hist, name, attr)

        self._unify_label_widths()
        return None

    def _fast_insert(self, key: int, obj: AbstractAction | Callable) -> None:
        """
        Insert object into the toolbar. Could be widget or callable.

        Parameters
        ----------
        key : int
            Position to insert.
        obj : Callable or AbstractAction
            Object to insert.
        """
        if isinstance(obj, Callable):
            # Sometimes users want to dynamically add new functions to GUI.
            if isinstance(obj, FunctionGui):
                if obj.parent is None:
                    f = nested_function_gui_callback(self, obj)
                    obj.called.connect(f)
                _obj = obj
            else:
                obj = _inject_recorder(obj, is_method=False).__get__(self)
                _obj = self._create_widget_from_method(obj)

            method_name = getattr(obj, "__name__", None)
            if method_name and not hasattr(self, method_name):
                object.__setattr__(self, method_name, obj)
        else:
            _obj = obj

        # _hide_labels should not contain Container because some ValueWidget like widgets
        # are Containers.
        if isinstance(_obj, self._component_class):
            insert_action_like(self.native, key, _obj.native)
            self._list.insert(key, _obj)

        elif isinstance(_obj, WidgetAction):
            if isinstance(_obj.widget, Separator):
                insert_action_like(self.native, key, "")

            else:
                _hide_labels = (
                    _LabeledWidgetAction,
                    ButtonWidget,
                    FreeWidget,
                    Label,
                    FunctionGui,
                    Image,
                    Table,
                )
                _obj = _obj
                if (not isinstance(_obj.widget, _hide_labels)) and self.labels:
                    _obj = _LabeledWidgetAction.from_action(_obj)
                _obj.parent = self
                insert_action_like(self.native, key, _obj.native)
            self._list.insert(key, _obj)
        else:
            raise TypeError(f"{type(_obj)} is not supported.")

    def insert(self, key: int, obj: AbstractAction) -> None:
        self._fast_insert(key, obj)
        self._unify_label_widths()


def _create_stylesheet(viewer: Viewer):
    if viewer is None:
        return ""
    w = viewer.window._qt_window
    styles = []
    for s in w.styleSheet().split("\n\n"):
        if s.startswith("QPushButton ") or s.startswith("QPushButton:"):
            styles.append("QToolButton" + s[11:])
    return "\n".join(styles)



from __future__ import annotations

from qtpy.QtCore import Qt
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent as _MouseClickEvent
from ._const import Modifier, Button


def _factorize_modifiers(mod):
    out = []
    if mod & Qt.ShiftModifier:
        out.append(Modifier.shift)
    if mod & Qt.ControlModifier:
        out.append(Modifier.control)
    if mod & Qt.AltModifier:
        out.append(Modifier.alt)
    return tuple(out)


def _factorize_buttons(mod):
    out = []
    if mod & Qt.LeftButton:
        out.append(Button.left)
    if mod & Qt.RightButton:
        out.append(Button.right)
    if mod & Qt.MiddleButton:
        out.append(Button.middle)
    return tuple(out)


class MouseClickEvent(_MouseClickEvent):
    def __init__(self, event: _MouseClickEvent, coord_item):
        self.accepted = event.accepted
        self.currentItem = (
            coord_item  # This enables mapping from event position to coordinates.
        )
        self._double = event._double
        self._scenePos = event._scenePos
        self._screenPos = event._screenPos
        self._button = event._button
        self._buttons = event._buttons
        self._modifiers = event._modifiers
        self._time = event._time
        self.acceptedItem = event.acceptedItem

    def pos(self):
        pos = super().pos()
        return (pos.x(), pos.y())

    def lastPos(self):
        pos = super().lastPos()
        return (pos.x(), pos.y())

    def modifiers(self):
        modifiers = super().modifiers()
        return _factorize_modifiers(modifiers)

    def buttons(self):
        buttons = super().buttons()
        return _factorize_buttons(buttons)



from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable, TypeVar
from types import FunctionType
from magicgui.widgets import FunctionGui, Widget
from magicgui.widgets._bases.value_widget import UNSET
from magicgui.type_map import get_widget_class
from magicgui.signature import magic_signature, MagicParameter, split_annotated_type
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    from ._base import BaseGui
    from .menu_gui import ContextMenuGui


def get_parameters(fgui: FunctionGui):
    return {k: v.default for k, v in fgui.__signature__.parameters.items()}


def set_context_menu(contextmenu: ContextMenuGui, parent: Widget | BaseGui) -> None:
    parent.native.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

    @parent.native.customContextMenuRequested.connect
    def rightClickContextMenu(point):
        contextmenu.native.exec_(parent.native.mapToGlobal(point))

    return None


_C = TypeVar("_C", bound=type)


def copy_class(cls: _C, ns: type, name: str | None = None) -> _C:
    """
    Copy a class in a new namespace.

    This function not only copies the class object but also update all the
    ``__qualname__`` recursively.

    Parameters
    ----------
    cls : type
        Class to be copied.
    ns : type
        New namespace of ``cls``.
    name : str, optional
        New name of ``cls``. If not given, the original name will be used.

    Returns
    -------
    type
        Copied class object.
    """
    out = type(cls.__name__, cls.__bases__, dict(cls.__dict__))
    if name is None:
        name = out.__name__
    _update_qualnames(out, f"{ns.__qualname__}.{name}")
    return out


def _update_qualnames(cls: type, cls_qualname: str) -> None:
    cls.__qualname__ = cls_qualname
    # NOTE: updating cls.__name__ will make `wraps` incompatible.
    for key, attr in cls.__dict__.items():
        if isinstance(attr, FunctionType):
            attr.__qualname__ = f"{cls_qualname}.{key}"
        elif isinstance(attr, type):
            _update_qualnames(attr, f"{cls_qualname}.{key}")

    return None


class MagicClassConstructionError(Exception):
    """Raised when class definition is not a valid magic-class."""


def format_error(
    e: Exception,
    hist: list[tuple[str, str, str]],
    name: str,
    attr: Any,
):
    hist_str = (
        "\n\t".join(map(lambda x: f"{x[0]} {x[1]} -> {x[2]}", hist))
        + f"\n\t\t{name} ({type(attr)}) <--- Error"
    )
    if not hist_str.startswith("\n\t"):
        hist_str = "\n\t" + hist_str
    if isinstance(e, MagicClassConstructionError):
        e.args = (f"\n{hist_str}\n{e}",)
        raise e
    else:
        tb = e.__traceback__
        construction_err = MagicClassConstructionError(
            f"\n{hist_str}\n\n{type(e).__name__}: {e}"
        ).with_traceback(tb)
        raise construction_err from None


def callable_to_classes(f: Callable) -> list[type[Widget]]:
    """Get list of classes that will be generated by the magicgui type map."""
    sig = magic_signature(f)
    return [_parameter_to_widget_class(p) for p in sig.parameters.values()]


TZ_EMPTY = "__no__default__"


def _parameter_to_widget_class(param: MagicParameter):
    value = UNSET if param.default in (param.empty, TZ_EMPTY) else param.default
    annotation, options = split_annotated_type(param.annotation)
    options = options.copy()
    wdg_class, _ = get_widget_class(value, annotation, options)
    return wdg_class



__version__ = "0.6.7"

from .core import (
    magicclass,
    magicmenu,
    magiccontext,
    magictoolbar,
    Parameters,
    build_help,
    get_function_gui,
    redo,
    update_widget_state,
)

from .wrappers import (
    set_options,
    set_design,
    do_not_record,
    bind_key,
    confirm,
    nogui,
    mark_preview,
)

from .fields import field, vfield, widget_property, FieldGroup, HasFields
from ._gui._base import wraps, defaults, MagicTemplate, PopUpMode
from ._gui.keybinding import Key
from ._gui._icon import Icon
from . import widgets, utils, types

from magicgui import *

__all__ = [
    "magicclass",
    "magicmenu",
    "magiccontext",
    "magictoolbar",
    "Parameters",
    "build_help",
    "get_function_gui",
    "redo",
    "update_widget_state",
    "set_options",
    "set_design",
    "do_not_record",
    "bind_key",
    "confirm",
    "nogui",
    "mark_preview",
    "field",
    "vfield",
    "widget_property",
    "FieldGroup",
    "HasFields",
    "wraps",
    "defaults",
    "MagicTemplate",
    "PopUpMode",
    "Key",
    "Icon",
]


def __getattr__(key: str):
    if key == "click":
        import warnings

        warnings.warn(
            "Function click is moving to magicclass.utils and will be deleted from "
            "magicclass namespace. Please 'from magicclass.utils import click'.",
            DeprecationWarning,
            stacklevel=2,
        )
        from .utils import click

        return click
    raise AttributeError(f"module {__name__!r} has no attribute {key!r}")



from .widgets2d import (
    VispyPlotCanvas,
    VispyImageCanvas,
    VispyMultiPlotCanvas,
    VispyMultiImageCanvas,
)
from .widgets3d import Vispy3DCanvas

__all__ = [
    "VispyPlotCanvas",
    "VispyImageCanvas",
    "VispyMultiPlotCanvas",
    "VispyMultiImageCanvas",
    "Vispy3DCanvas",
]



from __future__ import annotations
from dask.diagnostics import ResourceProfiler
from dask.diagnostics.profile import _Tracker, import_required, current_process
from timeit import default_timer
from time import sleep
import threading
from psygnal import Signal
from magicgui.widgets import Container, LineEdit
import datetime

from ...utils import QtSignal

# WIP!!


class DaskResourceProfiler(ResourceProfiler, Container):
    def __init__(self, dt: float = 1.0, **kwargs):
        ResourceProfiler.__init__(self, dt=dt)
        self._tic = LineEdit(value="0:00:00")
        self._mem = LineEdit(value="--- MB")
        self._cpu = LineEdit(value="--- %")
        Container.__init__(self, widgets=[self._tic, self._mem, self._cpu], **kwargs)

        self._running = True
        self._signal = QtSignal()
        self._signal.connect(self._update_display)

    def clear(self):
        ResourceProfiler.clear(self)

    def _start_collect(self):
        if not self._is_running():
            self._tracker = EventedTracker(self._dt)
            self._tracker.start()
            self._thread_timer = threading.Thread(target=self._thread_target)
            self._thread_timer.daemon = True
            self._thread_timer.start()

        self._tracker.parent_conn.send("collect")

    def _stop_collect(self):
        super()._stop_collect()
        self._running = False
        self._thread_timer.join()

    def _thread_target(self):
        while self._running:
            self._signal.emit(self._tracker._last_data)
            sleep(1)

    def _update_display(self, tp: tuple[float, float, float]):
        tic, mem, cpu = tp
        t = datetime.timedelta(seconds=int(tic))
        self._tic.value = str(t)
        self._mem.value = f"{mem} MB"
        self._cpu.value = f"{cpu} %"


class EventedTracker(_Tracker):
    _last_data = (0.0, 0.0, 0.0)

    def run(self):

        psutil = import_required(
            "psutil", "Tracking resource usage requires `psutil` to be installed"
        )
        self.parent = psutil.Process(self.parent_pid)

        pid = current_process()
        data = []
        while True:
            try:
                msg = self.child_conn.recv()
            except KeyboardInterrupt:
                continue
            if msg == "shutdown":
                break
            elif msg == "collect":
                ps = self._update_pids(pid)
                while not data or not self.child_conn.poll():
                    tic = default_timer()
                    mem = cpu = 0
                    for p in ps:
                        try:
                            mem2 = p.memory_info().rss
                            cpu2 = p.cpu_percent()
                        except Exception:  # could be a few different exceptions
                            pass
                        else:
                            # Only increment if both were successful
                            mem += mem2
                            cpu += cpu2
                    _new_data = (tic, mem / 1e6, cpu)
                    self._last_data = _new_data
                    data.append(_new_data)
                    sleep(self.dt)
            elif msg == "send_data":
                self.child_conn.send(data)
                data = []
        self.child_conn.close()



from ._functions import (
    iter_members,
    Tooltips,
    get_signature,
    argcount,
    is_instance_method,
    method_as_getter,
    eval_attribute,
    copy_info,
    show_tree,
    rst_to_html,
)

from .qt import (
    MessageBoxMode,
    show_messagebox,
    to_clipboard,
    open_url,
    screen_center,
    move_to_screen_center,
    screen_scale,
)

from ._click import click
from .qtsignal import QtSignal
from .qthreading import thread_worker, Timer, Callback



from __future__ import annotations
import numpy as np
from vispy import scene
from vispy.scene import visuals, ViewBox
from .layer2d import Curve, Scatter
from .._doc import write_docs
from ...widgets import FreeWidget


class HasViewBox(FreeWidget):
    def __init__(self, grid_pos: tuple[int, int] = (0, 0), _scene=None, _row=0, _col=0):
        if _scene is None:
            _scene = scene.SceneCanvas(keys="interactive")
        self._scene = _scene
        grid = self._scene.central_widget.add_grid(pos=grid_pos)
        grid.spacing = 0
        self._viewbox: ViewBox = grid.add_view(row=_row, col=_col, camera="panzoom")
        self._items = []
        self._grid = grid
        super().__init__()
        self._scene.create_native()
        self.set_widget(self._scene.native)

    @property
    def xrange(self) -> tuple[float, float]:
        return self._viewbox.camera._xlim

    @xrange.setter
    def xrange(self, rng: tuple[float, float]):
        x0, x1 = rng
        self._viewbox.camera.set_range(x=(x0, x1))

    @property
    def yrange(self) -> tuple[float, float]:
        return self._viewbox.camera._ylim

    @yrange.setter
    def yrange(self, rng: tuple[float, float]):
        y0, y1 = rng
        self._viewbox.camera.set_range(y=(y0, y1))

    @property
    def enabled(self) -> bool:
        return self._viewbox.interactive

    @enabled.setter
    def enabled(self, value) -> bool:
        self._viewbox.interactive = value

    @property
    def layers(self):
        return self._items

    @write_docs
    def add_curve(
        self,
        x=None,
        y=None,
        face_color=None,
        edge_color=None,
        color=None,
        size: float = 7,
        name: str | None = None,
        lw: float = 1,
        ls: str = "-",
        symbol=None,
    ):

        """
        Add a line plot like ``plt.plot(x, y)``.

        Parameters
        ----------
        {x}
        {y}
        {face_color}
        {edge_color}
        {color}
        size: float, default is 7
            Symbol size.
        {name}
        {lw}
        {ls}
        {symbol}

        Returns
        -------
        Curve
            A plot item of a curve.
        """
        x, y = _check_xy(x, y)
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        if isinstance(edge_color, np.ndarray) and edge_color.ndim == 1:
            edge_color = np.stack([edge_color] * y.size, axis=0)
        line = Curve(
            self._viewbox,
            x,
            y,
            face_color=face_color,
            edge_color=edge_color,
            size=size,
            name=name,
            lw=lw,
            ls=ls,
            symbol=symbol,
        )
        self._items.append(line)
        if len(self._items) == 1:
            self.xrange = (np.min(x), np.max(x))
            self.yrange = (np.min(y), np.max(y))
        return line

    @write_docs
    def add_scatter(
        self,
        x=None,
        y=None,
        face_color=None,
        edge_color=None,
        color=None,
        size: float = 7,
        name: str | None = None,
        symbol=None,
    ):

        """
        Add a line plot like ``plt.plot(x, y)``.

        Parameters
        ----------
        {x}
        {y}
        {face_color}
        {edge_color}
        {color}
        size: float, default is 7
            Symbol size.
        {name}
        {symbol}

        Returns
        -------
        Curve
            A plot item of a curve.
        """
        x, y = _check_xy(x, y)
        face_color, edge_color = _check_colors(face_color, edge_color, color)
        if isinstance(edge_color, np.ndarray) and edge_color.ndim == 1:
            edge_color = np.stack([edge_color] * y.size, axis=0)
        line = Scatter(
            self._viewbox,
            x,
            y,
            face_color=face_color,
            edge_color=edge_color,
            size=size,
            name=name,
            symbol=symbol,
        )
        self._items.append(line)
        if len(self._items) == 1:
            self.xrange = (np.min(x), np.max(x))
            self.yrange = (np.min(y), np.max(y))
        return line


class PlotItem(HasViewBox):
    def __init__(self, grid_pos=(0, 0), _scene=None):
        super().__init__(grid_pos=grid_pos, _scene=_scene, _row=1, _col=1)

        title = scene.Label("", color="white", font_size=7)
        title.height_max = 40
        self._grid.add_widget(title, row=0, col=0, col_span=2)
        self._title = title
        x_axis = scene.AxisWidget(
            orientation="bottom",
            anchors=("center", "bottom"),
            font_size=6,
            axis_label_margin=40,
            tick_label_margin=5,
            axis_label="",
        )
        x_axis.height_min = 65
        x_axis.height_max = 80
        x_axis.stretch = (1, 0.1)
        self._x_axis = x_axis
        self._grid.add_widget(x_axis, row=2, col=1)
        x_axis.link_view(self._viewbox)
        y_axis = scene.AxisWidget(
            orientation="left",
            anchors=("right", "middle"),
            font_size=6,
            axis_label_margin=50,
            tick_label_margin=5,
            axis_label="",
        )
        y_axis.width_max = 80
        y_axis.stretch = (0.1, 1)
        self._grid.add_widget(y_axis, row=1, col=0)
        y_axis.link_view(self._viewbox)
        self._y_axis = y_axis

    @property
    def title(self) -> str:
        """The title string."""
        return self._title.text

    @title.setter
    def title(self, text: str):
        self._title.text = text

    @property
    def xlabel(self) -> str:
        """The x-label string."""
        return self._x_axis.axis.axis_label

    @xlabel.setter
    def xlabel(self, text: str):
        self._x_axis.axis.axis_label = text
        height = self._x_axis.height
        if text:
            self._x_axis.size = (height, 75.0)
        else:
            self._x_axis.size = (height, 75.0)

    @property
    def ylabel(self) -> str:
        """The y-label string."""
        return self._y_axis.axis.axis_label

    @xlabel.setter
    def ylabel(self, text: str):
        self._y_axis.axis.axis_label = text


class ImageItem(HasViewBox):
    def __init__(
        self, lock_contrast_limits: bool = False, grid_pos=(0, 0), _scene=None
    ):
        super().__init__(grid_pos=grid_pos, _scene=_scene)
        self._viewbox.camera.aspect = 1.0
        self._viewbox.camera.flip = (False, True, False)
        self._image = visuals.Image(cmap="gray", parent=self._viewbox.scene)
        self._lock_contrast_limits = lock_contrast_limits

    @property
    def image(self):
        return self._image._data

    @image.setter
    def image(self, img):
        no_image = self._image._data is None
        self._image.set_data(img)
        if not self._lock_contrast_limits:
            self._image.clim = "auto"
        if no_image:
            self.yrange = (0, self._image._data.shape[0])
            self.xrange = (0, self._image._data.shape[1])

    @image.deleter
    def image(self):
        self._image._data = None
        self._image.update()

    @property
    def cmap(self):
        return self._image.cmap

    @cmap.setter
    def cmap(self, c):
        self._image.cmap = c


class VispyPlotCanvas(PlotItem):
    """A Vispy based 2-D plot canvas for curve, histogram, bar plot etc."""


class VispyImageCanvas(ImageItem):
    """A Vispy based 2-D plot canvas for images."""


class _MultiPlot(FreeWidget):
    _base_class: type[HasViewBox]

    def __init__(self, nrows: int = 1, ncols: int = 1):
        super().__init__()
        self._canvas = []
        self._scene = scene.SceneCanvas()
        for r in range(nrows):
            for c in range(ncols):
                canvas = self._base_class(grid_pos=(r, c), _scene=self._scene)
                self._canvas.append(canvas)

        self._scene.create_native()
        self.set_widget(self._scene.native)

    def __getitem__(self, i):
        return self._canvas[i]


class VispyMultiPlotCanvas(_MultiPlot):
    _base_class = VispyPlotCanvas


class VispyMultiImageCanvas(_MultiPlot):
    _base_class = VispyImageCanvas


def _check_xy(x, y):
    if y is None:
        if x is None:
            x = np.array([])
            y = np.array([])
        else:
            y = x
            x = np.arange(len(y))

    return x, y


def _check_colors(face_color, edge_color, color):
    if color is None:
        return face_color, edge_color
    else:
        if face_color is None and edge_color is None:
            return color, color
        else:
            raise ValueError(
                "Cannot set 'color' and either 'face_color' or "
                "'edge_color' at the same time."
            )



from __future__ import annotations
from typing import Callable
import warnings
from magicgui.widgets import Image, Table, Label, FunctionGui
from magicgui.widgets._bases import ButtonWidget
from magicgui.widgets._bases.widget import Widget
from macrokit import Symbol
from qtpy.QtWidgets import QMenu

from .mgui_ext import AbstractAction, WidgetAction, _LabeledWidgetAction
from .keybinding import register_shortcut
from ._base import (
    BaseGui,
    PopUpMode,
    ErrorMode,
    ContainerLikeGui,
    nested_function_gui_callback,
    _inject_recorder,
)
from .utils import copy_class, format_error

from ..signature import get_additional_option
from ..fields import MagicField
from ..widgets import Separator, FreeWidget
from ..utils import iter_members


def _check_popupmode(popup_mode: PopUpMode):
    if popup_mode in (PopUpMode.above, PopUpMode.below, PopUpMode.first):
        msg = (
            f"magicmenu does not support popup mode {popup_mode.value}."
            "PopUpMode.popup is used instead"
        )
        warnings.warn(msg, UserWarning)
    elif popup_mode == PopUpMode.last:
        msg = (
            f"magicmenu does not support popup mode {popup_mode.value}."
            "PopUpMode.parentlast is used instead"
        )
        warnings.warn(msg, UserWarning)
        popup_mode = PopUpMode.parentlast

    return popup_mode


class MenuGuiBase(ContainerLikeGui):
    def __init__(
        self,
        parent=None,
        name: str = None,
        close_on_run: bool = None,
        popup_mode: str | PopUpMode = None,
        error_mode: str | ErrorMode = None,
        labels: bool = True,
    ):
        popup_mode = _check_popupmode(popup_mode)

        super().__init__(
            close_on_run=close_on_run, popup_mode=popup_mode, error_mode=error_mode
        )
        name = name or self.__class__.__name__
        self.native = QMenu(name, parent)
        self.native.setToolTipsVisible(True)
        self.name = name
        self._list: list[MenuGuiBase | AbstractAction] = []
        self.labels = labels

    def _convert_attributes_into_widgets(self):
        cls = self.__class__

        # Add class docstring as label.
        if cls.__doc__:
            self.native.setToolTip(cls.__doc__)

        # Bind all the methods and annotations
        base_members = {x[0] for x in iter_members(MenuGuiBase)}

        _hist: list[tuple[str, str, str]] = []  # for traceback
        _ignore_types = (property, classmethod, staticmethod)

        for name, attr in filter(lambda x: x[0] not in base_members, iter_members(cls)):
            if isinstance(attr, _ignore_types):
                continue

            try:
                if isinstance(attr, type):
                    # Nested magic-menu
                    widget = attr()
                    object.__setattr__(self, name, widget)

                elif isinstance(attr, MagicField):
                    widget = self._create_widget_from_field(name, attr)

                else:
                    # convert class method into instance method
                    widget = getattr(self, name, None)

                if isinstance(widget, FunctionGui):
                    widget[0].bind(self)  # set self to the first argument

                elif isinstance(widget, BaseGui):
                    widget.__magicclass_parent__ = self
                    self.__magicclass_children__.append(widget)
                    widget._my_symbol = Symbol(name)

                    if isinstance(widget, MenuGuiBase):
                        widget.native.setParent(
                            self.native, widget.native.windowFlags()
                        )

                    else:
                        widget = WidgetAction(widget)

                elif isinstance(widget, Widget):
                    if not widget.name:
                        widget.name = name
                    if hasattr(widget, "text") and not widget.text:
                        widget.text = widget.name.replace("_", " ")
                    widget = WidgetAction(widget)

                if isinstance(widget, (MenuGuiBase, AbstractAction, Callable, Widget)):
                    if (not isinstance(widget, Widget)) and callable(widget):
                        if name.startswith("_") or not get_additional_option(
                            attr, "gui", True
                        ):
                            keybinding = get_additional_option(attr, "keybinding", None)
                            if keybinding:
                                register_shortcut(
                                    keys=keybinding, parent=self.native, target=widget
                                )
                            continue
                        widget = self._create_widget_from_method(widget)

                    elif hasattr(widget, "__magicclass_parent__") or hasattr(
                        widget.__class__, "__magicclass_parent__"
                    ):
                        if isinstance(widget, BaseGui):
                            widget._my_symbol = Symbol(name)
                        # magic-class has to know its parent.
                        # if __magicclass_parent__ is defined as a property, hasattr must be called
                        # with a type object (not instance).
                        widget.__magicclass_parent__ = self

                    if name.startswith("_"):
                        continue
                    moveto = get_additional_option(attr, "into")
                    copyto = get_additional_option(attr, "copyto", [])
                    if moveto is not None or copyto:
                        self._unwrap_method(name, widget, moveto, copyto)
                    else:
                        self._fast_insert(len(self), widget)

                    _hist.append((name, str(type(attr)), type(widget).__name__))

            except Exception as e:
                format_error(e, _hist, name, attr)

        self._unify_label_widths()
        return None

    def _fast_insert(
        self, key: int, obj: Callable | MenuGuiBase | AbstractAction
    ) -> None:
        if isinstance(obj, Callable):
            # Sometimes users want to dynamically add new functions to GUI.
            if isinstance(obj, FunctionGui):
                if obj.parent is None:
                    f = nested_function_gui_callback(self, obj)
                    obj.called.connect(f)
                _obj = obj
            else:
                obj = _inject_recorder(obj, is_method=False).__get__(self)
                _obj = self._create_widget_from_method(obj)

            method_name = getattr(obj, "__name__", None)
            if method_name and not hasattr(self, method_name):
                object.__setattr__(self, method_name, obj)
        else:
            _obj = obj

        if isinstance(_obj, (self._component_class, MenuGuiBase)):
            insert_action_like(self.native, key, _obj.native)
            self._list.insert(key, _obj)

        elif isinstance(_obj, WidgetAction):
            from .toolbar import ToolBarGui

            if isinstance(_obj.widget, Separator):
                insert_action_like(self.native, key, _obj.widget.title)

            elif isinstance(_obj.widget, ToolBarGui):
                qmenu = QMenu(_obj.widget.name, self.native)
                qmenu.addAction(_obj.native)
                if _obj.widget._icon_path is not None:
                    qmenu.setIcon(_obj.widget.native.windowIcon())
                insert_action_like(self.native, key, qmenu)

            else:
                _hide_labels = (
                    _LabeledWidgetAction,
                    ButtonWidget,
                    FreeWidget,
                    Label,
                    FunctionGui,
                    Image,
                    Table,
                )
                _obj = _obj
                if (not isinstance(_obj.widget, _hide_labels)) and self.labels:
                    _obj = _LabeledWidgetAction.from_action(_obj)
                _obj.parent = self
                insert_action_like(self.native, key, _obj.native)

            self._list.insert(key, _obj)
        else:
            raise TypeError(f"{type(_obj)} is not supported.")

    def insert(self, key: int, obj: Callable | MenuGuiBase | AbstractAction) -> None:
        """
        Insert object into the menu. Could be widget or callable.

        Parameters
        ----------
        key : int
            Position to insert.
        obj : Callable | MenuGuiBase | AbstractAction | Widget
            Object to insert.
        """
        self._fast_insert(key, obj)
        self._unify_label_widths()


def insert_action_like(qmenu: QMenu, key: int, obj):
    """
    Insert a QObject into a QMenu in a Pythonic way, like qmenu.insert(key, obj).

    Parameters
    ----------
    qmenu : QMenu
        QMenu object to which object will be inserted.
    key : int
        Position to insert.
    obj : QMenu or QAction or str
        Object to be inserted.
    """
    actions = qmenu.actions()
    l = len(actions)
    if key < 0:
        key = key + l
    if key == l:
        if isinstance(obj, QMenu):
            qmenu.addMenu(obj).setText(obj.objectName().replace("_", " "))
        elif isinstance(obj, str):
            if obj:
                qmenu.addSection(obj)
            else:
                qmenu.addSeparator()
        else:
            qmenu.addAction(obj)
    else:
        new_action = actions[key]
        before = new_action
        if isinstance(obj, QMenu):
            qmenu.insertMenu(before, obj).setText(obj.objectName().replace("_", " "))
        elif isinstance(obj, str):
            if obj:
                qmenu.insertSection(before, obj)
            else:
                qmenu.insertSeparator(before)
        else:
            qmenu.insertAction(before, obj)


class MenuGui(MenuGuiBase):
    """Magic class that will be converted into a menu bar."""


class ContextMenuGui(MenuGuiBase):
    """Magic class that will be converted into a context menu."""



from __future__ import annotations
from psygnal import Signal
from qtpy.QtWidgets import QFrame, QLabel, QPushButton, QSizePolicy, QHBoxLayout
from .utils import FreeWidget


class Separator(FreeWidget):
    """
    A Separator widget that can be used in both widgets and menus.
    This widget is not actually added to menus or toolbars.
    """

    btn_clicked = Signal(bool)

    def __init__(
        self,
        orientation: str = "horizontal",
        title: str = "",
        name: str = "",
        button: bool = False,
    ):
        super().__init__(name=name)
        self._qtitlebar = _QTitleBar(self.native, title, button)
        self.set_widget(self._qtitlebar)
        if button:
            self._qtitlebar.button.clicked.connect(
                lambda e: self.btn_clicked.emit(self._qtitlebar.button.isDown())
            )
        self._title = title

    @property
    def btn_text(self):
        return self._qtitlebar.button.text()

    @btn_text.setter
    def btn_text(self, value: str):
        self._qtitlebar.button.setText(value)

    @property
    def title(self) -> str:
        return self._title


class _QTitleBar(QLabel):
    """See also: napari/_qt/qt_main_window.py."""

    def __init__(self, parent, text: str = "", button: bool = False):
        super().__init__(parent)

        line = QFrame(parent=self)
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setObjectName("Separator")
        line.setStyleSheet(
            """
            QFrame#Separator {
                background-color: gray;
            }
        """
        )

        layout = QHBoxLayout()
        layout.setSpacing(4)
        layout.setContentsMargins(8, 1, 8, 0)

        if button:
            self.button = QPushButton(self)
            self.button.setFixedWidth(20)
            self.button.setStyleSheet(
                "QPushButton {"
                "border: 1px solid #555;"
                "border-radius: 10px;"
                "border-style: outset;"
                "padding: 5px"
                "}"
            )
            layout.addWidget(self.button)

        layout.addWidget(line)

        if text:
            text = QLabel(text, self)
            text.setSizePolicy(
                QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
            )
            layout.addWidget(text)

        self.setLayout(layout)

    def sizeHint(self):
        szh = super().sizeHint()
        szh.setHeight(20)
        return szh



from magicclass import magicclass, magicmenu

# PushButtonPlus.running should be True only if it is called from GUI.

def test_running():
    @magicclass
    class A:
        def __init__(self):
            self.last_running = None

        def func(self):
            self.last_running = self["func"].running

    ui = A()
    ui.func()
    assert ui.last_running == False
    ui["func"].changed()
    assert ui.last_running == True


def test_wrapped_running():
    @magicclass
    class A:
        @magicmenu
        class B:
            def func(self): ...

        def __init__(self):
            self.last_running = None

        @B.wraps
        def func(self):
            self.last_running = self["func"].running

    ui = A()
    ui.func()
    assert ui.last_running == False
    ui.B["func"].changed()
    assert ui.last_running == True



from __future__ import annotations
from macrokit import Symbol, Expr, Head, Macro, parse
from typing import TYPE_CHECKING, Iterable, overload
from qtpy.QtWidgets import QMenuBar, QMenu, QAction
from magicgui.widgets import FileEdit

from ..widgets.misc import FreeWidget, ConsoleTextEdit
from ..utils import to_clipboard, show_messagebox

if TYPE_CHECKING:
    from ._base import BaseGui


# TODO: Tabs


class MacroEdit(FreeWidget):
    """A text edit embeded with a custom menu bar."""

    count = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__magicclass_parent__ = None
        self.textedit = ConsoleTextEdit()
        self.set_widget(self.textedit.native)
        self.native.setWindowTitle("Macro")

        self._synchronize = True
        self._set_menubar()

    @property
    def value(self):
        """Get macro text."""
        return self.textedit.value

    @value.setter
    def value(self, value: str):
        self.textedit.value = value

    @property
    def synchronize(self):
        """Update macro text in real time if true."""
        return self._synchronize

    @synchronize.setter
    def synchronize(self, value: bool):
        if value and not self._synchronize:
            parent = self._search_parent_magicclass()
            parent.macro._last_setval = None  # To avoid overwriting the last code.
        self._synchronize = value

    def load(self, path: str):
        """Load macro text from a file."""
        path = str(path)
        with open(path) as f:
            self.value = f.read()

    def save(self, path: str):
        """Save current macro text."""
        path = str(path)
        with open(path, mode="w") as f:
            f.write(self.value)

    def _close(self, e=None):
        """Close widget."""
        return self.close()

    def execute(self):
        """Execute macro."""
        parent = self._search_parent_magicclass()
        with parent.macro.blocked():
            try:
                # substitute :ui and :viewer to the actual objects
                code = parse(self.textedit.value)
                _ui = Symbol.var("ui")
                code.eval({_ui: parent})
            except Exception as e:
                show_messagebox(
                    "error", title=e.__class__.__name__, text=str(e), parent=self.native
                )

    def execute_lines(self, line_numbers: int | slice | Iterable[int] = -1):
        """Execute macro at selected lines."""
        parent = self._search_parent_magicclass()
        with parent.macro.blocked():
            try:
                code = parse(self.textedit.value)
                if isinstance(line_numbers, (int, slice)):
                    lines = code.args[line_numbers]
                else:
                    lines = "\n".join(code.args[l] for l in line_numbers)
                lines.eval({Symbol.var("ui"): parent})
            except Exception as e:
                show_messagebox(
                    "error", title=e.__class__.__name__, text=str(e), parent=self.native
                )

    def _create_duplicate(self, e=None):
        self.duplicate().show()

    def _get_complete(self, e=None):
        if self.value:
            self = self.duplicate()
            self.show()
        parent = self._search_parent_magicclass()
        self.value = str(parent.macro)

    def _auto_pep8(self, e=None):
        import autopep8

        self.value = autopep8.fix_code(self.value)
        parent = self._search_parent_magicclass()
        parent.macro._last_setval = None  # To avoid overwriting the last code.

    def new(self, name: str = None) -> MacroEdit:
        """
        Create a new widget with same parent magic class widget.

        Parameters
        ----------
        name : str, optional
            Widget name. This name will be the title.

        Returns
        -------
        MacroEdit
            New MacroEdit widget.
        """
        if name is None:
            name = f"Macro-{self.__class__.count}"
            self.__class__.count += 1
        m = self.__class__(name=name)
        m.__magicclass_parent__ = self.__magicclass_parent__
        m.native.setParent(self.native.parent(), m.native.windowFlags())
        # Cannot synchronize sub widgets.
        m._synchronize = False
        m._synchronize_action.setChecked(False)
        m._synchronize_action.setEnabled(False)
        return m

    def duplicate(self, name=None) -> MacroEdit:
        """
        Create a new widget with same parent magic class widget. The code will be
        duplicated to the new one.

        Parameters
        ----------
        name : str, optional
            Widget name. This name will be the title.

        Returns
        -------
        MacroEdit
            New MacroEdit widget.
        """
        new = self.new(name=name)
        new.value = self.value
        return new

    def show(self):
        from ..utils import move_to_screen_center

        super().show()
        move_to_screen_center(self.native)

    def _execute(self, e=None):
        """Run macro."""
        self.execute()

    def _execute_selected(self, e=None):
        """Run selected line of macro."""
        parent = self._search_parent_magicclass()
        with parent.macro.blocked():
            try:
                all_code: str = self.textedit.value
                selected = self.textedit.selected
                code = parse(selected.strip())

                # to safely run code, every line should be fully selected even if selected
                # region does not raise SyntaxError.
                l = len(selected)
                start = all_code.find(selected)
                end = start + l
                if start != 0 and "\n" not in all_code[start - 1 : start + 1]:
                    raise SyntaxError("Select full line(s).")
                if end < l and "\n" not in all_code[end : end + 2]:
                    raise SyntaxError("Select full line(s).")

                code.eval({Symbol.var("ui"): parent})
            except Exception as e:
                show_messagebox(
                    "error", title=e.__class__.__name__, text=str(e), parent=self.native
                )

    def _search_parent_magicclass(self) -> BaseGui:
        current_self = self
        while getattr(current_self, "__magicclass_parent__", None) is not None:
            current_self = current_self.__magicclass_parent__
        return current_self

    def _load(self, e=None):
        """Load macro"""
        fdialog = FileEdit(mode="r", filter="*.txt;*.py")
        result = fdialog._show_file_dialog(
            fdialog.mode,
            caption=fdialog._btn_text,
            start_path=str(fdialog.value),
            filter=fdialog.filter,
        )
        if result:
            self.load(result)

    def _copy(self, e=None):
        """Copy text to clipboard"""
        to_clipboard(self.value)

    def _save(self, e=None):
        """Save text."""
        fdialog = FileEdit(mode="w", filter="*.txt;*.py")
        result = fdialog._show_file_dialog(
            fdialog.mode,
            caption=fdialog._btn_text,
            start_path=str(fdialog.value),
            filter=fdialog.filter,
        )
        if result:
            self.save(result)

    def _set_menubar(self):

        self._menubar = QMenuBar(self.native)
        self.native.layout().setMenuBar(self._menubar)

        self.textedit.read_only = False
        vbar = self.textedit.native.verticalScrollBar()
        vbar.setValue(vbar.maximum())

        # set file menu
        self._file_menu = QMenu("File", self.native)
        self._menubar.addMenu(self._file_menu)

        load_action = QAction("Load", self._file_menu)
        load_action.triggered.connect(self._load)
        self._file_menu.addAction(load_action)

        copy_action = QAction("Copy", self._file_menu)
        copy_action.triggered.connect(self._copy)
        self._file_menu.addAction(copy_action)

        save_action = QAction("Save", self._file_menu)
        save_action.triggered.connect(self._save)
        self._file_menu.addAction(save_action)

        close_action = QAction("Close", self._file_menu)
        close_action.triggered.connect(self._close)
        self._file_menu.addAction(close_action)

        # set macro menu
        self._macro_menu = QMenu("Macro", self.native)
        self._menubar.addMenu(self._macro_menu)

        run_action = QAction("Execute", self._macro_menu)
        run_action.triggered.connect(self._execute)
        self._macro_menu.addAction(run_action)

        run1_action = QAction("Execute selected lines", self._macro_menu)
        run1_action.triggered.connect(self._execute_selected)
        self._macro_menu.addAction(run1_action)

        create_action = QAction("Create", self._macro_menu)
        create_action.triggered.connect(self._create_duplicate)
        self._macro_menu.addAction(create_action)

        complete_action = QAction("Get complete macro", self._macro_menu)
        complete_action.triggered.connect(self._get_complete)
        self._macro_menu.addAction(complete_action)

        try:
            import autopep8

            pep8_action = QAction("Run PEP8", self._macro_menu)
            pep8_action.triggered.connect(self._auto_pep8)
            self._macro_menu.addAction(pep8_action)
        except ImportError:
            pass

        syn = QAction("Synchronize", self._macro_menu)
        syn.setCheckable(True)
        syn.setChecked(True)

        @syn.triggered.connect
        def _set_synchronize(check: bool):
            self.synchronize = check

        self._macro_menu.addAction(syn)
        self._synchronize_action = syn


class GuiMacro(Macro):
    def __init__(self, max_lines: int, flags={}):
        super().__init__(flags=flags)
        self._widget = None
        self._max_lines = max_lines
        self.callbacks.append(self._update_widget)

    @property
    def widget(self) -> MacroEdit:
        """Returns the macro editor."""
        if self._widget is None:
            self._widget = MacroEdit(name="Macro")
            from datetime import datetime

            now = datetime.now()
            self.append(Expr(Head.comment, [now.strftime("%Y/%m/%d %H:%M:%S")]))
        return self._widget

    def copy(self) -> Macro:
        """GuiMacro does not support deepcopy (and apparently _widget should not be copied)."""
        from copy import deepcopy

        return Macro(deepcopy(self.args), flags=self._flags)

    @overload
    def __getitem__(self, key: int) -> Expr:
        ...

    @overload
    def __getitem__(self, key: slice) -> Macro:
        ...

    def __getitem__(self, key):
        if isinstance(key, slice):
            return Macro(self._args, flags=self.flags)
        return super().__getitem__(key)

    def _update_widget(self, expr=None):
        if self.widget.synchronize:
            self.widget.textedit.append(str(self.args[-1]))
        if len(self) > self._max_lines:
            del self[0]

    def _erase_last(self):
        if self.widget.synchronize:
            self.widget.textedit.erase_last()



from __future__ import annotations
from typing import Any, TYPE_CHECKING, Callable, TypeVar
import re
from magicgui.widgets import PushButton
from magicgui.widgets._concrete import _LabeledWidget
from magicgui.widgets._bases import ValueWidget, ButtonWidget, ContainerWidget
from magicgui.widgets._function_gui import FunctionGui

from ..widgets import Separator

if TYPE_CHECKING:
    from magicgui.widgets import Widget


_R = TypeVar("_R")


class FunctionGuiPlus(FunctionGui[_R]):
    """FunctionGui class with a parameter recording functionality etc."""

    _dialog_widget = None

    def __call__(self, *args: Any, update_widget: bool = False, **kwargs: Any) -> _R:
        sig = self.__signature__
        try:
            bound = sig.bind(*args, **kwargs)
        except TypeError as e:
            if "missing a required argument" in str(e):
                match = re.search("argument: '(.+)'", str(e))
                missing = match.groups()[0] if match else "<param>"
                msg = (
                    f"{e} in call to '{self._callable_name}{sig}'.\n"
                    "To avoid this error, you can bind a value or callback to the "
                    f"parameter:\n\n    {self._callable_name}.{missing}.bind(value)"
                    "\n\nOr use the 'bind' option in the set_option decorator:\n\n"
                    f"    @set_option({missing}={{'bind': value}})\n"
                    f"    def {self._callable_name}{sig}: ..."
                )
                raise TypeError(msg) from None
            else:
                raise

        if update_widget:
            self._auto_call, before = False, self._auto_call
            try:
                self.update(bound.arguments)
            finally:
                self._auto_call = before

        bound.apply_defaults()

        # 1. Parameter recording
        # This is important when bound function set by {"bind": f} updates something.
        # When the value is referred via "__signature__" the bound function get called
        # and updated againg.
        self._previous_bound = bound

        self._tqdm_depth = 0  # reset the tqdm stack count

        # 2. Running flag
        # We sometimes want to know if the function is called programmatically or
        # from GUI. The "running" argument is True only when it's called via GUI.
        self.running = True
        try:
            value = self._function(*bound.args, **bound.kwargs)
        finally:
            self.running = False

        self._call_count += 1
        if self._result_widget is not None:
            with self._result_widget.changed.blocked():
                self._result_widget.value = value

        return_type = sig.return_annotation
        if return_type:
            from magicgui.type_map import _type2callback

            for callback in _type2callback(return_type):
                callback(self, value, return_type)
        self.called.emit(value)
        return value

    def insert(self, key: int, widget: Widget):
        """Insert widget at ``key``."""
        if isinstance(widget, (ValueWidget, ContainerWidget)):
            widget.changed.connect(lambda: self.changed.emit(self))
        _widget = widget

        if self.labels:
            # no labels for button widgets (push buttons, checkboxes, have their own)
            if not isinstance(widget, (_LabeledWidget, ButtonWidget, Separator)):
                _widget = _LabeledWidget(widget)
                widget.label_changed.connect(self._unify_label_widths)

        if key < 0:
            key += len(self)
        self._list.insert(key, widget)
        # NOTE: if someone has manually mucked around with self.native.layout()
        # it's possible that indices will be off.
        self._widget._mgui_insert_widget(key, _widget)
        self._unify_label_widths()

    def append_preview(self, f: Callable, text: str = "Preview"):
        """Append a preview button to the widget."""
        return append_preview(self, f, text)

    def exec_as_dialog(self, parent=None):
        """Show container as a dialog."""
        if self._dialog_widget is None:
            from magicgui.widgets import Dialog

            dlg = Dialog(widgets=[self], labels=False)
            self._dialog_widget = dlg
        else:
            dlg = self._dialog_widget
        dlg.native.setParent(parent.native, dlg.native.windowFlags())
        if self._dialog_widget.exec():
            self()
        return self._dialog_widget


def append_preview(self: FunctionGui, f: Callable, text: str = "Preview"):
    """Append a preview button to a FunctionGui widget."""

    btn = PushButton(text=text)
    if isinstance(self[-1], PushButton):
        self.insert(len(self) - 1, btn)
    else:
        self.append(btn)

    @btn.changed.connect
    def _call_preview():
        sig = self.__signature__
        bound = sig.bind()
        bound.apply_defaults()
        return f(*bound.args, **bound.kwargs)

    return f



from magicclass import (
    magicclass,
    magicmenu,
    magictoolbar,
    field,
    vfield,
    FieldGroup,
    HasFields,
    set_design,
)
from magicclass.fields import widget_property
from magicclass.types import Optional
from magicgui import widgets
from typing import Tuple
from unittest.mock import MagicMock
from pathlib import Path

def test_field_types():
    @magicclass
    class A:
        a_int = field(int)
        a_float = field(float)
        a_str = field(str)
        a_bool = field(bool)

    @magicclass
    class B:
        a_int = field(0)
        a_float = field(0.0)
        a_str = field("0")
        a_bool = field(False)

    @magicclass
    class C:
        a_int = field(widgets.SpinBox)
        a_float = field(widgets.FloatSpinBox)
        a_str = field(widgets.LineEdit)
        a_bool = field(widgets.CheckBox)

    @magicclass
    class D:
        a_int = field(widget_type=widgets.SpinBox)
        a_float = field(widget_type=widgets.FloatSpinBox)
        a_str = field(widget_type=widgets.LineEdit)
        a_bool = field(widget_type=widgets.CheckBox)

    @magicclass
    class E:
        a_int = field(widget_type="SpinBox")
        a_float = field(widget_type="FloatSpinBox")
        a_str = field(widget_type="LineEdit")
        a_bool = field(widget_type="CheckBox")

    answers = [widgets.SpinBox, widgets.FloatSpinBox,
               widgets.LineEdit, widgets.CheckBox]

    for cls in [A, B, C, D, E]:
        ui = cls()
        assert len(ui) == 4
        for i in range(4):
            widget = ui[i]
            assert type(widget) is answers[i]

    b = B()
    assert b.a_int.value == 0
    assert b.a_float.value == 0.0
    assert b.a_str.value == "0"
    assert b.a_bool.value == False

    b.a_int.value = 1
    len0 = len(b.macro)
    assert str(b.macro[-1]) == "ui.a_int.value = 1"
    b.a_int.value = 2
    len1 = len(b.macro)
    assert str(b.macro[-1]) == "ui.a_int.value = 2"
    assert len0 == len1


def test_vfield_types():
    @magicclass
    class A:
        a_int = vfield(int)
        a_float = vfield(float)
        a_str = vfield(str)
        a_bool = vfield(bool)

    @magicclass
    class B:
        a_int = vfield(0)
        a_float = vfield(0.0)
        a_str = vfield("0")
        a_bool = vfield(False)

    @magicclass
    class C:
        a_int = vfield(widgets.SpinBox)
        a_float = vfield(widgets.FloatSpinBox)
        a_str = vfield(widgets.LineEdit)
        a_bool = vfield(widgets.CheckBox)

    answers = [widgets.SpinBox, widgets.FloatSpinBox,
               widgets.LineEdit, widgets.CheckBox]

    for cls in [A, B, C]:
        ui = cls()
        assert len(ui) == 4
        for i in range(4):
            widget = ui[i]
            assert type(widget) is answers[i]

    b = B()
    assert b.a_int == 0
    assert b.a_float == 0.0
    assert b.a_str == "0"
    assert b.a_bool == False

    b.a_int = 1
    len0 = len(b.macro)
    assert str(b.macro[-1]) == "ui.a_int = 1"

    b.a_int = 2
    len1 = len(b.macro)
    assert str(b.macro[-1]) == "ui.a_int = 2"
    assert len0 == len1

    b.a_str = "x"
    assert str(b.macro[-2]) == "ui.a_int = 2"
    assert str(b.macro[-1]) == "ui.a_str = 'x'"


def test_field_options():
    tooltip = "this is int"
    name = "New Name"
    @magicclass
    class A:
        a_slider = field(1, widget_type=widgets.Slider)
        a_rename = field(1, name=name)
        a_minmax = field(1, options={"min": -10, "max": 10})
        a_tooltip = field(1, options={"tooltip": tooltip})

    ui = A()
    for i in range(4):
        assert ui[i].value == 1

    assert type(ui[0]) is widgets.Slider
    assert ui[1].name == name
    assert (ui[2].min, ui[2].max) == (-10, 10)
    ui[2].value = -5 # this should work
    assert ui[3].tooltip == tooltip


def test_fields_in_child():
    # test field
    @magicclass
    class A:
        @magicclass
        class B:
            def f1(self): ...
            i = field(str)
        x = field(int)

    ui = A()
    ui.x.value = 10
    assert str(ui.macro[-1]) == "ui.x.value = 10"
    ui.B.i.value = "aaa"
    assert str(ui.macro[-1]) == "ui.B.i.value = 'aaa'"

    # test vfield
    @magicclass
    class A:
        @magicclass
        class B:
            def f1(self): ...
            i = vfield(str)
        x = vfield(int)

    ui = A()
    ui.x = 10
    assert str(ui.macro[-1]) == "ui.x = 10"
    ui.B.i = "aaa"
    assert str(ui.macro[-1]) == "ui.B.i = 'aaa'"

    # test field
    @magicclass
    class A:
        @magicclass
        class B:
            @magictoolbar
            class Tool:
                a2 = field(int)
            a1 = field(int)
        a0 = field(int)

    ui = A()
    ui.a0.value = 10
    assert str(ui.macro[-1]) == "ui.a0.value = 10"
    ui.B.Tool.a2.value = 10
    assert str(ui.macro[-1]) == "ui.B.Tool.a2.value = 10"


    # test vfield
    @magicclass
    class A:
        @magicclass
        class B:
            @magictoolbar
            class Tool:
                a2 = vfield(int)
            a1 = vfield(int)
        a0 = vfield(int)

    ui = A()
    ui.a0 = 10
    assert str(ui.macro[-1]) == "ui.a0 = 10"
    ui.B.Tool.a2 = 10
    assert str(ui.macro[-1]) == "ui.B.Tool.a2 = 10"


def test_widget_actions():
    @magicclass
    class A:
        @magicmenu
        class B:
            a_int = field(0)
            a_float = field(0.0)
            a_int_sl = field(0, widget_type="Slider")
            a_float_sl = field(0, widget_type="FloatSlider")
            a_str = field("0")
            a_bool = field(False)

    ui = A()

    @magicclass
    class A:
        @magictoolbar
        class B:
            a_int = field(0)
            a_float = field(0.0)
            a_int_sl = field(0, widget_type="Slider")
            a_float_sl = field(0, widget_type="FloatSlider")
            a_str = field("0")
            a_bool = field(False)

    ui = A()

    @magicclass
    class A:
        @magicmenu
        class B:
            a_int = vfield(0)
            a_float = vfield(0.0)
            a_int_sl = vfield(0, widget_type="Slider")
            a_float_sl = vfield(0, widget_type="FloatSlider")
            a_str = vfield("0")
            a_bool = vfield(False)

    ui = A()

    @magicclass
    class A:
        @magictoolbar
        class B:
            a_int = vfield(0)
            a_float = vfield(0.0)
            a_int_sl = vfield(0, widget_type="Slider")
            a_float_sl = vfield(0, widget_type="FloatSlider")
            a_str = vfield("0")
            a_bool = vfield(False)

    ui = A()


def test_dont_record():
    @magicclass
    class A:
        t = field(int, record=True)
        f = field(int, record=False)

    ui = A()
    ui.t.value = 10
    assert str(ui.macro[-1]) == "ui.t.value = 10"
    ui.f.value = 10
    assert str(ui.macro[-1]) == "ui.t.value = 10"
    ui.t.value = 20
    assert str(ui.macro[-2]) != "ui.t.value = 10"
    assert str(ui.macro[-1]) == "ui.t.value = 20"

    @magicclass
    class A:
        t = vfield(int, record=True)
        f = vfield(int, record=False)

    ui = A()
    ui.t = 10
    assert str(ui.macro[-1]) == "ui.t = 10"
    ui.f = 10
    assert str(ui.macro[-1]) == "ui.t = 10"
    ui.t = 20
    assert str(ui.macro[-2]) != "ui.t = 10"
    assert str(ui.macro[-1]) == "ui.t = 20"

def test_enabled():
    @magicclass
    class A:
        @magicclass
        class B:
            a1 = field(int, options={"enabled": False})
            a2 = vfield(int, options={"enabled": False})
        b1 = field("b1", options={"enabled": False})
        b2 = field("b2", options={"enabled": False})

    ui = A()
    assert not ui[1].enabled
    assert not ui[2].enabled
    assert not ui.B[0].enabled
    assert not ui.B[1].enabled

def test_get_value_field_widget():
    @magicclass
    class A:
        x_1 = field(int)
        y_1 = vfield(int)
    ui = A()
    assert type(ui.y_1) is int
    assert type(ui["x_1"]) is widgets.SpinBox
    assert type(ui["y_1"]) is widgets.SpinBox

def test_generic_and_annotated():
    w0 = field(Tuple[int, str]).to_widget()
    w1 = field(Optional[int], options={"text": "XXX", "options": {"max": 10}}).to_widget()
    w2 = field(Optional[Tuple[int, str]]).to_widget()

    assert w0.widget_type == "TupleEdit"
    assert w1.widget_type == "OptionalWidget"
    assert w0[0].widget_type == "SpinBox"
    assert w0[1].widget_type == "LineEdit"
    assert w1[0].text == "XXX"
    assert w1[1].max == 10
    assert w2[1].widget_type == "TupleEdit"

def test_field_in_non_gui():
    """Test fields can be used in non-GUI classes."""
    class A:
        x = field(int)
        y = vfield(str)

        def __init__(self):
            self._x_value = None
            self._y_value = None

        @x.connect
        def _x(self):
            self._x_value = self.x.value

        @y.connect
        def _y(self):
            self._y_value = self.y

    a = A()
    assert a.x.widget_type == "SpinBox"
    a.x.value = 1
    assert a._x_value == 1
    a.y = "xxx"
    assert a._y_value == "xxx"

def test_field_group():
    from magicgui.widgets import Container

    class Params(FieldGroup):
        x = field(int)
        y = vfield(str)

    class A:
        params = Params(layout="horizontal")
        params2 = Params(layout="horizontal")

    a0 = A()
    assert isinstance(a0.params, Container)
    assert a0.params.x.widget_type == "SpinBox"
    assert a0.params.y == ""
    a0.params.x.value = 10
    a0.params.y = "t"
    assert a0.params2.x.value == 0
    assert a0.params2.y == ""


    a1 = A()
    assert a1.params.x.value == 0
    assert a1.params.y == ""

    @magicclass
    class Main:
        params = Params()

    ui = Main()
    assert ui.params is ui[0]


def test_nesting_field_group():
    class Params(FieldGroup):
        x = field(int)
        y = vfield(str)

    class G(FieldGroup):
        p = Params()
        u = field(bool)
        v = vfield(float)

    class A:
        g = G()

        def __init__(self):
            self.out = "None"
            self.g.p.signals.x.connect(lambda: self.set_output("x"))
            self.g.p.signals.y.connect(lambda: self.set_output("y"))

        def set_output(self, out):
            self.out = out

    a0 = A()
    a1 = A()

    assert a0.g.p.x.value == 0
    assert a0.g.p.y == ""
    assert a0.g.p is a0.g.p
    assert a0.g.widgets.p is a0.g.widgets.p
    assert a0.g.p.widgets.x is a0.g.p.widgets.x
    assert a0.g.p is not a1.g.p

    a0.g.p.x.value = 1
    assert a0.out == "x"
    assert a1.out == "None"
    a0.g.p.y = "a"
    assert a0.out == "y"
    assert a1.out == "None"
    a1.g.p.y = "xx"
    assert a0.out == "y"
    assert a1.out == "y"


def test_has_fields():
    class A(HasFields):
        x = vfield(int)
        y = vfield(str)
        @property
        def value(self):
            return self.x, self.y

    a0 = A()
    a1 = A()
    assert len(A._fields) == 2

    # test repr works
    repr(a0)
    repr(a0.widgets)
    repr(a0.signals)

    a0.x = 10
    a0.y = "abc"

    assert (10, "abc") == a0.value
    assert (0, "") == a1.value

    c0 = a0.widgets.as_container()
    assert len(c0) == 2
    assert c0[0].name == "x"
    assert c0[1].name == "y"
    assert c0["x"].value == a0.x
    assert c0["y"].value == a0.y

def test_widget_property():
    from magicgui.widgets import Slider
    mock = MagicMock()

    class A(HasFields):
        def __init__(self, max=5):
            self._max = max
            self.result = None

        @widget_property
        def a(self):
            return Slider(max=self._max)

        @a.connect
        def _a(self, v):
            self.result = v

    x = A(max=10)
    y = A(max=20)

    assert x.widgets is not y.widgets
    assert x.widgets.a.max == 10
    assert y.widgets.a.max == 20

    x.a = 1
    assert x.result == 1
    assert y.result is None

def test_tooltip():
    @magicclass
    class A:
        """
        Test class.

        Attributes
        ----------
        a : SpinBox
            Parameter-a.
        b : str
            Parameter-b.
        """
        a = field(int)
        b = vfield(str)

    ui = A()
    assert ui.a.tooltip == "Parameter-a."
    assert ui["b"].tooltip == "Parameter-b."

    class B(HasFields):
        """
        Test class.

        Attributes
        ----------
        a : SpinBox
            Parameter-a.
        b : str
            Parameter-b.
        """
        a = field(int)
        b = vfield(str)

    ui = B()
    assert ui.widgets.a.tooltip == "Parameter-a."
    assert ui.widgets.b.tooltip == "Parameter-b."

def test_get_set_hooks():
    class A:
        offset = 1
        suffix = "-0"

        x = vfield(int)
        y = vfield(str)

        @x.pre_set_hook
        def _x_set(self, value):
            return value + self.offset

        @y.post_get_hook
        def _y_get(self, value):
            return value + self.suffix

    a = A()
    a.x = 10
    assert a.x == 11
    a.y = "Y"
    assert a.y == "Y-0"



from __future__ import annotations
import sys
from typing import (
    TYPE_CHECKING,
    Generic,
    Iterable,
    MutableSequence,
    Any,
    TypeVar,
)
from typing_extensions import _AnnotatedAlias, get_args
from psygnal import Signal
from qtpy import QtWidgets as QtW
from qtpy.QtGui import QTextCursor
from qtpy.QtCore import Qt
from magicgui.signature import split_annotated_type
from magicgui.widgets import (
    PushButton,
    TextEdit,
    Table,
    Container,
    CheckBox,
    create_widget,
)
from magicgui.application import use_app
from magicgui.widgets import LineEdit
from magicgui.widgets._bases.value_widget import ValueWidget, UNSET
from magicgui.backends._qtpy.widgets import (
    QBaseWidget,
    LineEdit as BaseLineEdit,
)
from .utils import FreeWidget, merge_super_sigs

if TYPE_CHECKING:
    from qtpy.QtWidgets import QTextEdit
    from superqt import QLabeledRangeSlider


if sys.platform == "win32":
    _FONT = "Consolas"
else:
    _FONT = "Menlo"


@merge_super_sigs
class OptionalWidget(Container):
    """
    A container that can represent optional argument.

    Parameters
    ----------
    widget_type : ValueWidget type
        Type of inner value widget.
    text : str, optional
        Text of checkbox.
    value : Any
        Initial value.
    options : dict, optional
        Widget options of the inner value widget.
    """

    def __init__(
        self,
        inner_widget: type[ValueWidget] | None = None,
        text: str = None,
        layout="vertical",
        nullable=True,
        value=UNSET,
        options=None,
        **kwargs,
    ):
        if text is None:
            text = "Use default value"
        if options is None:
            options = {}
        self._checkbox = CheckBox(text=text, value=True)

        if inner_widget is None:
            annot = kwargs.get("annotation", None)
            if annot is None:
                annot_arg = type(value)
            else:
                args = get_args(annot)
                if len(args) > 0:
                    annot_arg = args[0]
                else:
                    annot_arg = type(value)

            if isinstance(annot_arg, _AnnotatedAlias):
                annot_arg, metadata = split_annotated_type(annot_arg)
                options.update(metadata)

            self._inner_value_widget = create_widget(
                annotation=annot_arg,
                options=options,
            )

        else:
            self._inner_value_widget = inner_widget

        super().__init__(
            layout=layout,
            widgets=(self._checkbox, self._inner_value_widget),
            labels=True,
            **kwargs,
        )

        @self._checkbox.changed.connect
        def _toggle_visibility(v: bool):
            self._inner_value_widget.visible = not v

        self.value = value

    @property
    def value(self) -> Any:
        if not self._checkbox.value:
            return self._inner_value_widget.value
        else:
            return None

    @value.setter
    def value(self, v: Any) -> None:
        if v is None or v is UNSET:
            self._checkbox.value = True
            self._inner_value_widget.visible = False
        else:
            self._inner_value_widget.value = v
            self._checkbox.value = False
            self._inner_value_widget.visible = True

    @property
    def text(self) -> str:
        return self._checkbox.text

    @text.setter
    def text(self, v: str) -> None:
        self._checkbox.text = v


class ConsoleTextEdit(TextEdit):
    """A text edit with console-like setting."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from qtpy.QtGui import QFont, QTextOption

        self.native: QTextEdit
        font = QFont(_FONT)
        font.setStyleHint(QFont.Monospace)
        font.setFixedPitch(True)
        self.native.setFont(font)
        self.native.setWordWrapMode(QTextOption.NoWrap)

        # set tab width
        self.tab_size = 4

    @property
    def tab_size(self):
        metrics = self.native.fontMetrics()
        return self.native.tabStopWidth() // metrics.width(" ")

    @tab_size.setter
    def tab_size(self, size: int):
        metrics = self.native.fontMetrics()
        self.native.setTabStopWidth(size * metrics.width(" "))

    def append(self, text: str):
        """Append new text."""
        self.native.append(text)

    def erase_last(self):
        """Erase the last line."""
        cursor = self.native.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.select(QTextCursor.LineUnderCursor)
        cursor.removeSelectedText()
        cursor.deletePreviousChar()
        self.native.setTextCursor(cursor)

    def erase_first(self):
        """Erase the first line."""
        cursor = self.native.textCursor()
        cursor.movePosition(QTextCursor.Start)
        cursor.select(QTextCursor.LineUnderCursor)
        cursor.removeSelectedText()
        cursor.movePosition(QTextCursor.Down)
        cursor.deletePreviousChar()
        cursor.movePosition(QTextCursor.End)
        self.native.setTextCursor(cursor)

    @property
    def selected(self) -> str:
        """Return selected string."""
        cursor = self.native.textCursor()
        return cursor.selectedText().replace("\u2029", "\n")


class CheckButton(PushButton):
    """A checkable button."""

    def __init__(self, text: str | None = None, **kwargs):
        super().__init__(text=text, **kwargs)
        self.native.setCheckable(True)


class QIntEdit(BaseLineEdit):
    _qwidget: QtW.QLineEdit

    def _post_get_hook(self, value):
        if value == "":
            return None
        return int(value)

    def _pre_set_hook(self, value):
        return str(value)


class IntEdit(LineEdit):
    def __init__(self, value=UNSET, **kwargs):
        app = use_app()
        assert app.native
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=QIntEdit,
            **kwargs,
        )


class QFloatEdit(BaseLineEdit):
    _qwidget: QtW.QLineEdit

    def _post_get_hook(self, value):
        if value == "":
            return None
        return float(value)

    def _pre_set_hook(self, value):
        return str(value)


class FloatEdit(LineEdit):
    def __init__(self, value=UNSET, **kwargs):
        app = use_app()
        assert app.native
        ValueWidget.__init__(
            self,
            value=value,
            widget_type=QFloatEdit,
            **kwargs,
        )


_V = TypeVar("_V")


class QRangeSlider(QBaseWidget):
    _qwidget: QLabeledRangeSlider

    def _mgui_get_value(self):
        pass

    def _mgui_bind_change_callback(self, callback):
        pass

    def _mgui_set_value(self, rng):
        pass


class AbstractRangeSlider(ValueWidget, Generic[_V]):
    """
    A slider widget that represent a range like (2, 5).

    This class is a temporary one and may be substituted by magicgui widget soon.
    See https://github.com/napari/magicgui/pull/337.
    """

    changed = Signal(tuple)

    def __init__(
        self,
        value=UNSET,
        min=0,
        max=1000,
        orientation: str = "horizontal",
        nullable: bool = True,
        **kwargs,
    ):
        sl = self._construct_qt()
        sl.setMinimum(min)
        sl.setMaximum(max)
        sl.valueChanged.connect(self.changed)
        if orientation == "horizontal":
            sl.setOrientation(Qt.Horizontal)
        elif orientation == "vertical":
            sl.setOrientation(Qt.Vertical)
        else:
            raise ValueError(
                "Only horizontal and vertical orientation are currently supported"
            )
        self._slider = sl
        super().__init__(
            value=value,
            widget_type=QRangeSlider,
            backend_kwargs={"qwidg": QtW.QWidget},
            **kwargs,
        )
        self.native.setLayout(QtW.QVBoxLayout())
        self.native.setContentsMargins(0, 0, 0, 0)
        self.native.layout().addWidget(sl)

    @classmethod
    def _construct_qt(cls, *args, **kwargs) -> QLabeledRangeSlider:
        raise NotImplementedError()

    @property
    def value(self) -> tuple[_V, _V]:
        return self._slider.value()

    @value.setter
    def value(self, rng: tuple[_V, _V]) -> None:
        x0, x1 = rng
        if x0 > x1:
            raise ValueError(f"lower value exceeds higher value ({x0} > {x1}).")
        self._slider.setValue((x0, x1))

    @property
    def range(self) -> tuple[_V, _V]:
        return self._slider.minimum(), self._slider.maximum()

    @range.setter
    def range(self, rng: tuple[_V, _V]) -> None:
        x0, x1 = rng
        if x0 > x1:
            raise ValueError(f"Minimum value exceeds maximum value ({x0} > {x1}).")
        self._slider.setMinimum(x0)
        self._slider.setMaximum(x1)

    @property
    def min(self) -> _V:
        return self._slider.minimum()

    @min.setter
    def min(self, value: _V) -> None:
        self._slider.setMinimum(value)

    @property
    def max(self) -> _V:
        return self._slider.maximum()

    @max.setter
    def max(self, value: _V) -> None:
        self._slider.setMaximum(value)


class RangeSlider(AbstractRangeSlider[int]):
    @classmethod
    def _construct_qt(cls, *args, **kwargs):
        from superqt import QLabeledRangeSlider

        sl = QLabeledRangeSlider()
        sl.setHandleLabelPosition(QLabeledRangeSlider.LabelPosition.LabelsAbove)
        sl.setEdgeLabelMode(QLabeledRangeSlider.EdgeLabelMode.NoLabel)
        return sl


class FloatRangeSlider(AbstractRangeSlider[float]):
    @classmethod
    def _construct_qt(cls, *args, **kwargs):
        from superqt import QLabeledDoubleRangeSlider

        sl = QLabeledDoubleRangeSlider()
        sl.setHandleLabelPosition(QLabeledDoubleRangeSlider.LabelPosition.LabelsAbove)
        sl.setEdgeLabelMode(QLabeledDoubleRangeSlider.EdgeLabelMode.NoLabel)
        return sl


class _QtSpreadSheet(QtW.QTabWidget):
    def __init__(self):
        super().__init__()
        self.setMovable(True)
        self._n_table = 0
        self.tabBar().tabBarDoubleClicked.connect(self.editTabBarLabel)
        self.tabBar().setContextMenuPolicy(Qt.CustomContextMenu)
        self.tabBar().customContextMenuRequested.connect(self.showContextMenu)
        self._line_edit = None

    def addTable(self, table):
        self.addTab(table, f"Sheet {self._n_table}")
        self._n_table += 1

    def renameTab(self, index: int, name: str) -> None:
        self.tabBar().setTabText(index, name)
        return None

    def editTabBarLabel(self, index: int):
        if index < 0:
            return
        if self._line_edit is not None:
            self._line_edit.deleteLater()
            self._line_edit = None

        tabbar = self.tabBar()
        self._line_edit = QtW.QLineEdit(self)

        @self._line_edit.editingFinished.connect
        def _(_=None):
            self.renameTab(index, self._line_edit.text())
            self._line_edit.deleteLater()
            self._line_edit = None

        self._line_edit.setText(tabbar.tabText(index))
        self._line_edit.setGeometry(tabbar.tabRect(index))
        self._line_edit.setFocus()
        self._line_edit.selectAll()
        self._line_edit.show()

    def showContextMenu(self, point):
        if point.isNull():
            return
        tabbar = self.tabBar()
        index = tabbar.tabAt(point)
        menu = QtW.QMenu(self)
        rename_action = menu.addAction("Rename")
        rename_action.triggered.connect(lambda _: self.editTabBarLabel(index))
        delete_action = menu.addAction("Delete")
        delete_action.triggered.connect(lambda _: self.removeTab(index))

        menu.exec(tabbar.mapToGlobal(point))


class SpreadSheet(FreeWidget, MutableSequence[Table]):
    """A simple spread sheet widget."""

    def __init__(self, read_only: bool = False):
        super().__init__()
        spreadsheet = _QtSpreadSheet()
        self.set_widget(spreadsheet)
        self.central_widget: _QtSpreadSheet
        self._tables: list[Table] = []
        self.read_only = read_only

    def __len__(self) -> int:
        return self.central_widget.count()

    def index(self, item: Table | str):
        if isinstance(item, Table):
            for i, table in enumerate(self._tables):
                if item is table:
                    return i
            else:
                raise ValueError
        elif isinstance(item, str):
            tabbar = self.central_widget.tabBar()
            for i in range(tabbar.count()):
                text = tabbar.tabText(i)
                if text == item:
                    return i
            else:
                raise ValueError
        else:
            raise TypeError

    def __getitem__(self, key):
        if isinstance(key, str):
            key = self.index(key)
        return self._tables[key]

    def __setitem__(self, key, value):
        raise NotImplementedError

    def __delitem__(self, key):
        if isinstance(key, str):
            key = self.index(key)
        self.central_widget.removeTab(key)
        del self._tables[key]

    def __iter__(self) -> Iterable[Table]:
        return iter(self._tables)

    def insert(self, key: int, value):
        """Insert a table-like data as a new sheet."""
        if key < 0:
            key += len(self)
        table = Table(value=value)
        table.read_only = self.read_only
        self.central_widget.addTable(table.native)
        self._tables.insert(key, table)
        return None

    def rename(self, index: int, name: str):
        """Rename tab at index `index` with name `name`."""
        self.central_widget.renameTab(index, name)
        return None

    @property
    def read_only(self):
        return self._read_only

    @read_only.setter
    def read_only(self, v: bool):
        for table in self._tables:
            table.read_only = v
        self._read_only = v



from __future__ import annotations
from functools import wraps as functools_wraps
import inspect
from weakref import WeakValueDictionary
from typing import Any, TYPE_CHECKING

from ._gui.class_gui import (
    ClassGuiBase,
    ClassGui,
    FrameClassGui,
    GroupBoxClassGui,
    MainWindowClassGui,
    SubWindowsClassGui,
    ScrollableClassGui,
    DraggableClassGui,
    ButtonClassGui,
    CollapsibleClassGui,
    HCollapsibleClassGui,
    SplitClassGui,
    TabbedClassGui,
    StackedClassGui,
    ToolBoxClassGui,
    ListClassGui,
)
from ._gui._base import (
    PopUpMode,
    ErrorMode,
    defaults,
    MagicTemplate,
    check_override,
    convert_attributes,
)
from ._gui import ContextMenuGui, MenuGui, ToolBarGui
from ._app import get_app
from .types import WidgetType
from . import _register  # activate type registration things.

if TYPE_CHECKING:
    from .stylesheets import StyleSheet
    from ._gui import MenuGuiBase
    from ._gui._function_gui import FunctionGuiPlus
    from .types import WidgetTypeStr, PopUpModeStr, ErrorModeStr
    from .help import HelpWidget
    from macrokit import Macro

_BASE_CLASS_SUFFIX = "_Base"

_TYPE_MAP = {
    WidgetType.none: ClassGui,
    WidgetType.scrollable: ScrollableClassGui,
    WidgetType.draggable: DraggableClassGui,
    WidgetType.split: SplitClassGui,
    WidgetType.collapsible: CollapsibleClassGui,
    WidgetType.hcollapsible: HCollapsibleClassGui,
    WidgetType.button: ButtonClassGui,
    WidgetType.toolbox: ToolBoxClassGui,
    WidgetType.tabbed: TabbedClassGui,
    WidgetType.stacked: StackedClassGui,
    WidgetType.list: ListClassGui,
    WidgetType.groupbox: GroupBoxClassGui,
    WidgetType.frame: FrameClassGui,
    WidgetType.subwindows: SubWindowsClassGui,
    WidgetType.mainwindow: MainWindowClassGui,
}


def magicclass(
    class_: type | None = None,
    *,
    layout: str = "vertical",
    labels: bool = True,
    name: str = None,
    visible: bool | None = None,
    close_on_run: bool = None,
    popup_mode: PopUpModeStr | PopUpMode = None,
    error_mode: ErrorModeStr | ErrorMode = None,
    widget_type: WidgetTypeStr | WidgetType = WidgetType.none,
    icon_path: str = None,
    stylesheet: str | StyleSheet = None,
):
    """
    Decorator that can convert a Python class into a widget.

    .. code-block:: python

        @magicclass
        class C:
            ...
        ui = C()
        ui.show()  # open GUI

    Parameters
    ----------
    class_ : type, optional
        Class to be decorated.
    layout : str, "vertical" or "horizontal", default is "vertical"
        Layout of the main widget.
    labels : bool, default is True
        If true, magicgui labels are shown.
    name : str, optional
        Name of GUI.
    visible : bool, optional
        Initial visibility of GUI. Useful when magic class is nested.
    close_on_run : bool, default is True
        If True, magicgui created by every method will be deleted after the method is completed without
        exceptions, i.e. magicgui is more like a dialog.
    popup : bool, default is True
        Deprecated.
    popup_mode : str or PopUpMode, default is PopUpMode.popup
        Option of how to popup FunctionGui widget when a button is clicked.
    error_mode : str or ErrorMode, default is ErrorMode.msgbox
        Option of how to raise errors during function calls.
    widget_type : WidgetType or str, optional
        Widget type of container.
    icon_path : str, optional
        Path to the icon image.
    stylesheet : str or StyleSheet object, optional
        Set stylesheet to the widget if given.

    Returns
    -------
    Decorated class or decorator.
    """
    if popup_mode is None:
        popup_mode = defaults["popup_mode"]
    if close_on_run is None:
        close_on_run = defaults["close_on_run"]
    if error_mode is None:
        error_mode = defaults["error_mode"]

    if isinstance(widget_type, str):
        widget_type = widget_type.lower()

    widget_type = WidgetType(widget_type)

    def wrapper(cls) -> type[ClassGui]:
        if not isinstance(cls, type):
            raise TypeError(f"magicclass can only wrap classes, not {type(cls)}")

        class_gui = _TYPE_MAP[widget_type]

        if not issubclass(cls, MagicTemplate):
            check_override(cls)

        # get class attributes first
        doc = cls.__doc__
        sig = inspect.signature(cls)
        annot = cls.__dict__.get("__annotations__", {})
        mod = cls.__module__
        qualname = cls.__qualname__

        new_attrs = convert_attributes(cls, hide=class_gui.__mro__)
        oldclass = type(cls.__name__ + _BASE_CLASS_SUFFIX, (cls,), {})
        newclass = type(cls.__name__, (class_gui, oldclass), new_attrs)

        newclass.__signature__ = sig
        newclass.__doc__ = doc
        newclass.__module__ = mod
        newclass.__qualname__ = qualname

        # concatenate annotations
        newclass.__annotations__ = class_gui.__annotations__.copy()
        newclass.__annotations__.update(annot)

        @functools_wraps(oldclass.__init__)
        def __init__(self: MagicTemplate, *args, **kwargs):
            # Without "app = " Jupyter freezes after closing the window!
            app = get_app()

            gui_kwargs = dict(
                layout=layout,
                labels=labels,
                name=name or cls.__name__,
                visible=visible,
                close_on_run=close_on_run,
                popup_mode=PopUpMode(popup_mode),
                error_mode=ErrorMode(error_mode),
            )

            # Inheriting Container's constructor is the most intuitive way.
            if kwargs and "__init__" not in cls.__dict__:
                gui_kwargs.update(kwargs)
                kwargs = {}

            class_gui.__init__(
                self,
                **gui_kwargs,
            )
            # prepare macro
            macrowidget = self.macro.widget.native
            macrowidget.setParent(self.native, macrowidget.windowFlags())
            self.macro.widget.__magicclass_parent__ = self

            with self.macro.blocked():
                super(oldclass, self).__init__(*args, **kwargs)

            self._convert_attributes_into_widgets()

            if widget_type in (WidgetType.collapsible, WidgetType.button):
                self.text = self.name

            if icon_path:
                self.icon_path = icon_path
            if stylesheet:
                self.native.setStyleSheet(str(stylesheet))
            if hasattr(self, "__post_init__"):
                with self.macro.blocked():
                    self.__post_init__()

        newclass.__init__ = __init__

        # Users may want to override repr
        newclass.__repr__ = oldclass.__repr__

        return newclass

    if class_ is None:
        return wrapper
    else:
        return wrapper(class_)


def magicmenu(
    class_: type = None,
    *,
    close_on_run: bool = None,
    popup_mode: str | PopUpMode = None,
    error_mode: str | ErrorMode = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str = None,
):
    """
    Decorator that can convert a Python class into a menu bar.
    """
    return _call_magicmenu(**locals(), menugui_class=MenuGui)


def magiccontext(
    class_: type = None,
    *,
    close_on_run: bool = None,
    popup_mode: str | PopUpMode = None,
    error_mode: str | ErrorMode = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str = None,
):
    """
    Decorator that can convert a Python class into a context menu.
    """
    return _call_magicmenu(**locals(), menugui_class=ContextMenuGui)


def magictoolbar(
    class_: type = None,
    *,
    close_on_run: bool = None,
    popup_mode: str | PopUpMode = None,
    error_mode: str | ErrorMode = None,
    labels: bool = True,
    name: str | None = None,
    icon_path: str = None,
):
    """
    Decorator that can convert a Python class into a menu bar.
    """
    return _call_magicmenu(**locals(), menugui_class=ToolBarGui)


class MagicClassFactory:
    """
    Factory class that can make any magic-class.
    """

    def __init__(
        self,
        name: str,
        layout: str = "vertical",
        labels: bool = True,
        close_on_run: bool = None,
        popup_mode: str | PopUpMode = None,
        error_mode: str | ErrorMode = None,
        widget_type: str | WidgetType = WidgetType.none,
        icon_path: str = None,
        attrs: dict[str, Any] | None = None,
    ):
        self.name = name
        self.layout = layout
        self.labels = labels
        self.close_on_run = close_on_run
        self.popup_mode = popup_mode
        self.error_mode = error_mode
        self.widget_type = widget_type
        self.icon_path = icon_path
        self.attrs = attrs

    def as_magicclass(self) -> ClassGuiBase:
        _cls = type(self.name, (), self.attrs)
        cls = magicclass(
            _cls,
            layout=self.layout,
            labels=self.labels,
            close_on_run=self.close_on_run,
            name=self.name,
            popup_mode=self.popup_mode,
            error_mode=self.error_mode,
            widget_type=self.widget_type,
            icon_path=self.icon_path,
        )
        return cls

    def as_magictoolbar(self) -> ToolBarGui:
        _cls = type(self.name, (), self.attrs)
        cls = magictoolbar(
            _cls,
            close_on_run=self.close_on_run,
            popup_mode=self.popup_mode,
            error_mode=self.error_mode,
            labels=self.labels,
            icon_path=self.icon_path,
        )
        return cls

    def as_magicmenu(self) -> MenuGui:
        _cls = type(self.name, (), self.attrs)
        cls = magicmenu(
            _cls,
            close_on_run=self.close_on_run,
            popup_mode=self.popup_mode,
            error_mode=self.error_mode,
            labels=self.labels,
            icon_path=self.icon_path,
        )
        return cls

    def as_magiccontext(self) -> ContextMenuGui:
        _cls = type(self.name, (), self.attrs)
        cls = magiccontext(
            _cls,
            close_on_run=self.close_on_run,
            popup_mode=self.popup_mode,
            error_mode=self.error_mode,
            labels=self.labels,
            icon_path=self.icon_path,
        )
        return cls


def _call_magicmenu(
    class_: type = None,
    close_on_run: bool = True,
    popup_mode: str | PopUpMode = None,
    error_mode: str | ErrorMode = None,
    labels: bool = True,
    name: str = None,
    icon_path: str = None,
    menugui_class: type[MenuGuiBase] = None,
):
    """
    Parameters
    ----------
    class_ : type, optional
        Class to be decorated.
    close_on_run : bool, default is True
        If True, magicgui created by every method will be deleted after the method is completed without
        exceptions, i.e. magicgui is more like a dialog.
    popup : bool, default is True
        If True, magicgui created by every method will be poped up, else they will be appended as a
        part of the main widget.

    Returns
    -------
    Decorated class or decorator.
    """

    if popup_mode is None:
        popup_mode = defaults["popup_mode"]
    if close_on_run is None:
        close_on_run = defaults["close_on_run"]
    if error_mode is None:
        error_mode = defaults["error_mode"]

    if popup_mode in (
        PopUpMode.above,
        PopUpMode.below,
        PopUpMode.first,
        PopUpMode.last,
    ):
        raise ValueError(f"Mode {popup_mode.value} is not compatible with Menu.")

    def wrapper(cls) -> type[menugui_class]:
        if not isinstance(cls, type):
            raise TypeError(f"magicclass can only wrap classes, not {type(cls)}")

        if not issubclass(cls, MagicTemplate):
            check_override(cls)

        # get class attributes first
        doc = cls.__doc__
        sig = inspect.signature(cls)
        mod = cls.__module__
        qualname = cls.__qualname__

        new_attrs = convert_attributes(cls, hide=menugui_class.__mro__)
        oldclass = type(cls.__name__ + _BASE_CLASS_SUFFIX, (cls,), {})
        newclass = type(cls.__name__, (menugui_class, oldclass), new_attrs)

        newclass.__signature__ = sig
        newclass.__doc__ = doc
        newclass.__module__ = mod
        newclass.__qualname__ = qualname

        @functools_wraps(oldclass.__init__)
        def __init__(self: MagicTemplate, *args, **kwargs):
            # Without "app = " Jupyter freezes after closing the window!
            app = get_app()

            gui_kwargs = dict(
                close_on_run=close_on_run,
                popup_mode=PopUpMode(popup_mode),
                error_mode=ErrorMode(error_mode),
                labels=labels,
                name=name or cls.__name__,
            )

            # Inheriting Container's constructor is the most intuitive way.
            if kwargs and "__init__" not in cls.__dict__:
                gui_kwargs.update(kwargs)
                kwargs = {}

            menugui_class.__init__(
                self,
                **gui_kwargs,
            )

            macrowidget = self.macro.widget.native
            macrowidget.setParent(self.native, macrowidget.windowFlags())
            self.macro.widget.__magicclass_parent__ = self

            with self.macro.blocked():
                super(oldclass, self).__init__(*args, **kwargs)

            self._convert_attributes_into_widgets()

            if icon_path:
                self.icon_path = icon_path
            if hasattr(self, "__post_init__"):
                with self.macro.blocked():
                    self.__post_init__()

        newclass.__init__ = __init__

        # Users may want to override repr
        newclass.__repr__ = oldclass.__repr__

        return newclass

    return wrapper if class_ is None else wrapper(class_)


magicmenu.__doc__ += _call_magicmenu.__doc__
magiccontext.__doc__ += _call_magicmenu.__doc__

_HELPS: WeakValueDictionary[int, MagicTemplate] = WeakValueDictionary()


def build_help(ui: MagicTemplate, parent=None) -> HelpWidget:
    """
    Build a widget for user guide. Once it is built, widget will be cached.

    Parameters
    ----------
    ui : MagicTemplate
        Magic class UI object.

    Returns
    -------
    HelpWidget
        Help of the input UI.
    """
    ui_id = id(ui)
    if ui_id in _HELPS.keys():
        help_widget = _HELPS[ui_id]
    else:
        from .help import HelpWidget

        if parent is None:
            parent = ui.native
        help_widget = HelpWidget(ui, parent=parent)
        _HELPS[ui_id] = help_widget
    return help_widget


def get_function_gui(ui: MagicTemplate, name: str) -> FunctionGuiPlus:
    """
    Get the FunctionGui object hidden beneath push button or menu.

    This function is a helper function for magicclass.

    Parameters
    ----------
    ui : MagicTemplate
        Any of a magic-class instance.
    name : str
        Name of method (or strictly speaking, the name of PushButton).

    Returns
    -------
    FunctionGuiPlus
        FunctionGui object.
    """
    func = getattr(ui, name)
    widget = ui[name]

    if not hasattr(widget, "mgui"):
        raise TypeError(f"Widget {widget} does not have FunctionGui inside it.")

    if widget.mgui is not None:
        return widget.mgui

    from ._gui._base import _build_mgui, _create_gui_method

    func = _create_gui_method(ui, func)
    mgui = _build_mgui(widget, func, ui)
    return mgui


def redo(ui: MagicTemplate, index: int = -1) -> None:
    """
    Redo operation on GUI using recorded macro.

    Parameters
    ----------
    ui : MagicTemplate
        Target magic-class widget.
    index : int, default is -1
        Which execution will be redone. Any object that support list slicing can be used.
        By default the last operation will be redone.
    """
    line = ui.macro[index]
    try:
        line.eval({"ui": ui})
    except Exception as e:
        msg = e.args[0]
        msg = f"Caused by >>> {line}. {msg}"
        raise e
    return None


def update_widget_state(ui: MagicTemplate, macro: Macro | str | None = None) -> None:
    """
    Update widget values based on a macro.

    This helper function works similar to the ``update_widget`` method of ``FunctionGui``.
    In most cases, this function will be used for restoring a state from a macro recorded
    before. Value changed signal will not be emitted within this operation.

    Parameters
    ----------
    ui : MagicTemplate
        Magic class instance.
    macro : Macro or str, optional
        An executable macro or string that dictates how GUI will be updated.
    """
    from macrokit import Head, Expr, Macro

    if macro is None:
        macro = ui.macro
    elif isinstance(macro, str):
        s = macro
        macro = Macro()
        for line in s.split("\n"):
            macro.append(line)
    elif not isinstance(macro, Macro):
        raise TypeError(
            f"The second argument must be a Macro or str, got {type(macro)}."
        )

    for expr in macro:
        if expr.head == Head.call:
            # ui.func(...)
            ui_f, *arguments = expr.args
            fname = str(ui_f.args[1])
            if fname.startswith("_"):
                if fname != "_call_with_return_callback":
                    continue
                args, kwargs = _arguments_to_values(arguments)
                fgui = get_function_gui(ui, args[0])

            else:
                args, kwargs = _arguments_to_values(arguments)
                fgui = get_function_gui(ui, fname)

            with fgui.changed.blocked():
                for key, value in kwargs.items():
                    getattr(fgui, key).value = value

        elif expr.head == Head.assign:
            # ui.field.value = ...
            # ui.vfield = ...
            expr.eval({}, {str(ui._my_symbol): ui})

    return None


def _tuple(*args) -> tuple:
    return args


def _arguments_to_values(arguments) -> tuple[tuple, dict[str, Any]]:
    from macrokit import Head, Expr, symbol

    for i, arg in enumerate(arguments):
        if isinstance(arg, Expr) and arg.head == Head.kw:
            break
    args = arguments[:i]
    kwargs: list[Expr] = arguments[i:]
    args = Expr(Head.call, [_tuple] + args).eval({symbol(_tuple): _tuple})
    kwargs = Expr(Head.call, [dict] + kwargs).eval()
    return args, kwargs


class Parameters:
    def __init__(self):
        self.__name__ = self.__class__.__name__
        self.__qualname__ = self.__class__.__qualname__

        sig = [
            inspect.Parameter(name="self", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
        ]
        for name, attr in inspect.getmembers(self):
            if name.startswith("__") or callable(attr):
                continue
            sig.append(
                inspect.Parameter(
                    name=name,
                    kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    default=attr,
                )
            )
        if hasattr(self.__class__, "__annotations__"):
            annot = self.__class__.__annotations__
            for name, t in annot.items():
                sig.append(
                    inspect.Parameter(
                        name=name,
                        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        annotation=t,
                    )
                )

        self.__signature__ = inspect.Signature(sig)

    def __call__(self, *args) -> None:
        params = list(self.__signature__.parameters.keys())[1:]
        for a, param in zip(args, params):
            setattr(self, param, a)

    def as_dict(self) -> dict[str, Any]:
        """
        Convert parameter fields into a dictionary.

        .. code-block:: python

            class params(Parameters):
                i = 1
                j = 2

            p = params()
            p.as_dict() # {"i": 1, "j": 2}

        """
        params = list(self.__signature__.parameters.keys())[1:]
        return {param: getattr(self, param) for param in params}



from qtpy import QtCore
from typing import Callable, Any


class QtSignal(QtCore.QObject):
    """Dummy qt object for pyqt signal operations."""

    signal = QtCore.Signal(object)

    def connect(self, slot: Callable):
        return self.signal.connect(slot)

    def emit(self, val: Any = None):
        return self.signal.emit(val)



from typing import Tuple
from magicclass import magicclass, set_options, field, vfield, magictoolbar
from magicclass.types import Bound, Optional

def test_bind_value():
    # test binding an invariable
    @magicclass
    class A:
        @set_options(x={"bind": 10})
        def func(self, x):
            pass

    ui = A()
    ui["func"].changed()
    assert str(ui.macro[-1]) == "ui.func(x=10)"

def test_bind_method():
    @magicclass
    class A:
        def __init__(self):
            self.i = 0

        def _get_param(self, w=None):
            self.i += 1
            return self.i

        @set_options(x={"bind": _get_param})
        def func(self, x):
            pass

    ui0 = A()
    ui1 = A()

    ui0["func"].changed()
    assert str(ui0.macro[-1]) == "ui.func(x=1)"
    ui0["func"].changed()
    assert str(ui0.macro[-1]) == "ui.func(x=2)"
    assert ui0.i == 2
    assert ui1.i == 0

def test_bind_field():
    @magicclass
    class A:
        a = field(int)
        b = vfield(int)

        @set_options(a={"bind": a})
        def func_a(self, a):
            pass

        @set_options(b={"bind": b})
        def func_b(self, b):
            pass

    ui = A()
    ui["func_a"].changed()
    assert str(ui.macro[-1]) == "ui.func_a(a=0)"
    ui.a.value = 10
    ui["func_a"].changed()
    assert str(ui.macro[-1]) == "ui.func_a(a=10)"

    ui["func_b"].changed()
    assert str(ui.macro[-1]) == "ui.func_b(b=0)"
    ui.b = 10
    ui["func_b"].changed()
    assert str(ui.macro[-1]) == "ui.func_b(b=10)"

def test_Bound():
    @magicclass
    class A:
        def __init__(self):
            self.i = 0

        def _get_param(self, w=None):
            self.i += 1
            return self.i

        a = field("a")

        def func(self, x: Bound[_get_param], y: Bound[a]):
            return x, y

    ui = A()
    ui["func"].changed()
    assert str(ui.macro[-1]) == "ui.func(x=1, y='a')"

def test_wrapped():
    @magicclass
    class A:
        @magictoolbar
        class B:
            field_b = field("b")
            def _b(self, w=None):
                return "b"

            def func_2(self): ...

        field_a = field("a")
        def _a(self, w=None):
            return "a"

        def func_1(self, x: Bound[_a], y: Bound[B.field_b]):
            self.returned = x, y

        @B.wraps
        def func_2(self, x: Bound[field_a], y: Bound[B._b]):
            self.returned = x, y

    ui = A()
    ui["func_1"].changed()
    ui["func_2"].changed()
    assert str(ui.macro[-2]) == "ui.func_1(x='a', y='b')"
    assert str(ui.macro[-1]) == "ui.func_2(x='a', y='b')"

def test_nesting():
    @magicclass
    class A:
        @magicclass
        class B:
            def __init__(self):
                self._a = 0

            def _get_value(self, w=None):
                return self._a

            def func(self, x: Bound[_get_value]):
                self.returned = x

        def func(self, x: Bound[B._get_value]):
            self.returned = x

    ui = A()
    ui["func"].changed()
    ui.B["func"].changed()
    assert ui.returned == 0
    assert ui.B.returned == 0
    ui.B._a = 1
    ui["func"].changed()
    ui.B["func"].changed()
    assert ui.returned == 1
    assert ui.B.returned == 1

def test_external_field():
    @magicclass
    class B:
        x = field(int)

    @magicclass
    class A:
        _a = -1
        b = B
        @b.x.connect
        def _callback(self):
            self._a = 1

        def func(self, x: Bound[b.x]):
            self._a = x

    ui = A()
    assert ui._a == -1
    ui.b.x.value = 10
    assert ui._a == 1
    ui["func"].changed()
    assert ui._a == 10

def test_multi_gui():
    @magicclass
    class A:
        def _bind(self, w=None):
            return id(self)

        def f(self, a: Bound[_bind]):
            self._a = a

    a0 = A()
    a1 = A()

    a0["f"].changed()
    a1["f"].changed()

    assert a0._a == id(a0)
    assert a1._a == id(a1)

def test_annotated():
    @magicclass
    class A:
        x = field(Optional[float])
        y = field(Tuple[int, str])
        _out = None
        def f(self, x0: Bound[x], y0: Bound[y]):
            self._out = (x0, y0)

    ui = A()
    ui.x.value = 0.5
    ui.y.value = (1, "z")
    ui["f"].changed()
    assert ui._out is not None
    assert ui._out == (0.5, (1, "z"))

def test_bind_with_string():
    @magicclass
    class A:
        _out = None
        def f(self, x0: Bound["x"], y0: Bound["_get"]):
            self._out = (x0, y0)
        x = field(3)

        def _get(self, w=None):
            return 0

        def g(self, x0: Bound["B.bx"], y0: Bound["B._b_get"]):
            self._out = (x0, y0)

        @magicclass
        class B:
            bx = field(4)
            def _b_get(self, w=None):
                return 1

    ui = A()
    ui["f"].changed()
    assert ui._out == (3, 0)
    ui["g"].changed()
    assert ui._out == (4, 1)



from __future__ import annotations
from functools import wraps
from typing import Callable, Iterable, Union, TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from .._gui import BaseGui
    from .._gui.mgui_ext import Action
    from magicgui.widgets._bases import ButtonWidget

nStrings = Union[str, Iterable[str]]


def click(
    enables: nStrings | None = None,
    disables: nStrings | None = None,
    enabled: bool = True,
    shows: nStrings | None = None,
    hides: nStrings | None = None,
    visible: bool = True,
):
    """
    Set options of push buttons related to button clickability.

    Parameters
    ----------
    enables : str or iterable of str, optional
        Enables other button(s) in this list when clicked.
    disables : str or iterable of str, optional
        Disables other button(s) in this list when clicked.
    enabled : bool, default is True
        The initial clickability state of the button.
    shows : str or iterable of str, optional
        Make other button(s) in this list visible when clicked.
    hides : str or iterable of str, optional
        Make other button(s) in this list invisible when clicked.
    visible: bool, default is True
        The initial visibility of the button.
    """
    enables = _assert_iterable(enables)
    disables = _assert_iterable(disables)
    shows = _assert_iterable(shows)
    hides = _assert_iterable(hides)

    def wrapper(func):
        @wraps(func)
        def f(self, *args, **kwargs):
            out = func(self, *args, **kwargs)
            for button in _iter_widgets(self, enables):
                button.enabled = True
            for button in _iter_widgets(self, disables):
                button.enabled = False
            for button in _iter_widgets(self, shows):
                button.visible = True
            for button in _iter_widgets(self, hides):
                button.visible = False

            return out

        caller_options = {"enabled": enabled, "visible": visible}
        from ..signature import upgrade_signature

        upgrade_signature(f, caller_options=caller_options)
        return f

    return wrapper


def _assert_iterable(obj):
    if obj is None:
        obj = []
    elif isinstance(obj, str) or callable(obj):
        obj = [obj]
    return obj


def _iter_widgets(
    self: BaseGui, descriptors: Iterable[list[str]] | Iterable[Callable]
) -> Iterator[ButtonWidget | Action]:
    for f in descriptors:
        if callable(f):
            # A.B.func -> B.func, if self is an object of A.
            f = f.__qualname__.split(self.__class__.__name__)[1][1:]

        if isinstance(f, str):
            *clsnames, funcname = f.split(".")
            # search for parent class that match the description.
            ins = self
            for a in clsnames:
                if a != "":
                    ins = getattr(ins, a)
                else:
                    ins = ins.__magicclass_parent__

            button = ins[funcname]
        else:
            raise TypeError(f"Unexpected type in click decorator: {type(f)}")
        yield button



from __future__ import annotations
import os
from types import SimpleNamespace
from typing import TYPE_CHECKING, Any
from pathlib import Path
from qtpy.QtWidgets import QStyle, QApplication
from qtpy.QtGui import QIcon, QImage, QPixmap
from qtpy.QtCore import Qt, QSize

if TYPE_CHECKING:
    from .mgui_ext import PushButtonPlus, AbstractAction


class _IconBase:
    _source: Any

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._source!r})"

    def install(self, dst: PushButtonPlus | AbstractAction) -> None:
        raise NotImplementedError()


class StandardIcon(_IconBase):
    """An object of a standard icon."""

    def __init__(self, source: Any):
        if isinstance(source, str):
            source = getattr(Icon, source)
        self._source = source

    def install(self, dst: PushButtonPlus | AbstractAction) -> None:
        icon = QApplication.style().standardIcon(self._source)
        dst.native.setIcon(icon)


class IconPath(_IconBase):
    """An object of an icon from a path."""

    def __init__(self, source: Any):
        self._source = str(source)

    def __str__(self) -> str:
        return self._source

    def install(self, dst: PushButtonPlus | AbstractAction) -> None:
        icon = QIcon(self._source)
        dst.native.setIcon(icon)
        return None


class ArrayIcon(_IconBase):
    """An object of an icon from numpy array."""

    _source: QImage

    def __init__(self, source: Any):
        import numpy as np

        arr = np.asarray(source)

        from magicgui import _mpl_image

        img = _mpl_image.Image()

        img.set_data(arr)

        val: np.ndarray = img.make_image()
        h, w, _ = val.shape
        self._source = QImage(val, w, h, QImage.Format.Format_RGBA8888)

    def install(self, dst: PushButtonPlus | AbstractAction):
        if hasattr(dst.native, "size"):
            qsize = dst.native.size()
        else:
            qsize = QSize(32, 32)
        qimg = self._source.scaled(
            qsize,
            Qt.AspectRatioMode.KeepAspectRatio,
            Qt.TransformationMode.SmoothTransformation,
        )

        qpix = QPixmap.fromImage(qimg)
        icon = QIcon(qpix)
        dst.native.setIcon(icon)
        return None


def get_icon(val: Any) -> _IconBase:
    """Get a proper icon object from a value."""
    if isinstance(val, _IconBase):
        icon = val
    elif isinstance(val, int):
        icon = StandardIcon(val)
    elif isinstance(val, Path) or os.path.exists(val):
        icon = IconPath(val)
    elif hasattr(val, "__array__"):
        icon = ArrayIcon(val)
    elif isinstance(val, str) and hasattr(Icon, val):
        icon = StandardIcon(val)
    else:
        raise TypeError(f"Input {val!r} cannot be converted to an icon.")
    return icon


class Icon(SimpleNamespace):
    """
    Namespace of icons.

    .. code-block:: python

        from magicclass import magicclass, set_design, Icon

        @magicclass
        class A:
            @set_design(icon=Icon.FileIcon)
            def func(self):
                ...

    """

    # fmt: off
    TitleBarMenuButton = QStyle.StandardPixmap.SP_TitleBarMenuButton
    TitleBarMinButton = QStyle.StandardPixmap.SP_TitleBarMinButton
    TitleBarMaxButton = QStyle.StandardPixmap.SP_TitleBarMaxButton
    TitleBarCloseButton = QStyle.StandardPixmap.SP_TitleBarCloseButton
    TitleBarNormalButton = QStyle.StandardPixmap.SP_TitleBarNormalButton
    TitleBarShadeButton = QStyle.StandardPixmap.SP_TitleBarShadeButton
    TitleBarUnshadeButton = QStyle.StandardPixmap.SP_TitleBarUnshadeButton
    TitleBarContextHelpButton = QStyle.StandardPixmap.SP_TitleBarContextHelpButton
    MessageBoxInformation = QStyle.StandardPixmap.SP_MessageBoxInformation
    MessageBoxWarning = QStyle.StandardPixmap.SP_MessageBoxWarning
    MessageBoxCritical = QStyle.StandardPixmap.SP_MessageBoxCritical
    MessageBoxQuestion = QStyle.StandardPixmap.SP_MessageBoxQuestion
    DockWidgetCloseButton = QStyle.StandardPixmap.SP_DockWidgetCloseButton
    DesktopIcon = QStyle.StandardPixmap.SP_DesktopIcon
    TrashIcon = QStyle.StandardPixmap.SP_TrashIcon
    ComputerIcon = QStyle.StandardPixmap.SP_ComputerIcon
    DriveFDIcon = QStyle.StandardPixmap.SP_DriveFDIcon
    DriveHDIcon = QStyle.StandardPixmap.SP_DriveHDIcon
    DriveCDIcon = QStyle.StandardPixmap.SP_DriveCDIcon
    DriveDVDIcon = QStyle.StandardPixmap.SP_DriveDVDIcon
    DriveNetIcon = QStyle.StandardPixmap.SP_DriveNetIcon
    DirOpenIcon = QStyle.StandardPixmap.SP_DirOpenIcon
    DirClosedIcon = QStyle.StandardPixmap.SP_DirClosedIcon
    DirLinkIcon = QStyle.StandardPixmap.SP_DirLinkIcon
    FileIcon = QStyle.StandardPixmap.SP_FileIcon
    FileLinkIcon = QStyle.StandardPixmap.SP_FileLinkIcon
    ToolBarHorizontalExtensionButton = QStyle.StandardPixmap.SP_ToolBarHorizontalExtensionButton
    ToolBarVerticalExtensionButton = QStyle.StandardPixmap.SP_ToolBarVerticalExtensionButton
    FileDialogStart = QStyle.StandardPixmap.SP_FileDialogStart
    FileDialogEnd = QStyle.StandardPixmap.SP_FileDialogEnd
    FileDialogToParent = QStyle.StandardPixmap.SP_FileDialogToParent
    FileDialogNewFolder = QStyle.StandardPixmap.SP_FileDialogNewFolder
    FileDialogDetailedView = QStyle.StandardPixmap.SP_FileDialogDetailedView
    FileDialogInfoView = QStyle.StandardPixmap.SP_FileDialogInfoView
    FileDialogContentsView = QStyle.StandardPixmap.SP_FileDialogContentsView
    FileDialogListView = QStyle.StandardPixmap.SP_FileDialogListView
    FileDialogBack = QStyle.StandardPixmap.SP_FileDialogBack
    DirIcon = QStyle.StandardPixmap.SP_DirIcon
    DialogOkButton = QStyle.StandardPixmap.SP_DialogOkButton
    DialogCancelButton = QStyle.StandardPixmap.SP_DialogCancelButton
    DialogHelpButton = QStyle.StandardPixmap.SP_DialogHelpButton
    DialogOpenButton = QStyle.StandardPixmap.SP_DialogOpenButton
    DialogSaveButton = QStyle.StandardPixmap.SP_DialogSaveButton
    DialogCloseButton = QStyle.StandardPixmap.SP_DialogCloseButton
    DialogApplyButton = QStyle.StandardPixmap.SP_DialogApplyButton
    DialogResetButton = QStyle.StandardPixmap.SP_DialogResetButton
    DialogDiscardButton = QStyle.StandardPixmap.SP_DialogDiscardButton
    DialogYesButton = QStyle.StandardPixmap.SP_DialogYesButton
    DialogNoButton = QStyle.StandardPixmap.SP_DialogNoButton
    ArrowUp = QStyle.StandardPixmap.SP_ArrowUp
    ArrowDown = QStyle.StandardPixmap.SP_ArrowDown
    ArrowLeft = QStyle.StandardPixmap.SP_ArrowLeft
    ArrowRight = QStyle.StandardPixmap.SP_ArrowRight
    ArrowBack = QStyle.StandardPixmap.SP_ArrowBack
    ArrowForward = QStyle.StandardPixmap.SP_ArrowForward
    DirHomeIcon = QStyle.StandardPixmap.SP_DirHomeIcon
    CommandLink = QStyle.StandardPixmap.SP_CommandLink
    VistaShield = QStyle.StandardPixmap.SP_VistaShield
    BrowserReload = QStyle.StandardPixmap.SP_BrowserReload
    BrowserStop = QStyle.StandardPixmap.SP_BrowserStop
    MediaPlay = QStyle.StandardPixmap.SP_MediaPlay
    MediaStop = QStyle.StandardPixmap.SP_MediaStop
    MediaPause = QStyle.StandardPixmap.SP_MediaPause
    MediaSkipForward = QStyle.StandardPixmap.SP_MediaSkipForward
    MediaSkipBackward = QStyle.StandardPixmap.SP_MediaSkipBackward
    MediaSeekForward = QStyle.StandardPixmap.SP_MediaSeekForward
    MediaSeekBackward = QStyle.StandardPixmap.SP_MediaSeekBackward
    MediaVolume = QStyle.StandardPixmap.SP_MediaVolume
    MediaVolumeMuted = QStyle.StandardPixmap.SP_MediaVolumeMuted
    DirLinkOpenIcon = QStyle.StandardPixmap.SP_DirLinkOpenIcon
    LineEditClearButton = QStyle.StandardPixmap.SP_LineEditClearButton
    DialogYesToAllButton = QStyle.StandardPixmap.SP_DialogYesToAllButton
    DialogNoToAllButton = QStyle.StandardPixmap.SP_DialogNoToAllButton
    DialogSaveAllButton = QStyle.StandardPixmap.SP_DialogSaveAllButton
    DialogAbortButton = QStyle.StandardPixmap.SP_DialogAbortButton
    DialogRetryButton = QStyle.StandardPixmap.SP_DialogRetryButton
    DialogIgnoreButton = QStyle.StandardPixmap.SP_DialogIgnoreButton
    RestoreDefaultsButton = QStyle.StandardPixmap.SP_RestoreDefaultsButton
    CustomBase = QStyle.StandardPixmap.SP_CustomBase
    # fmt: on



from __future__ import annotations
from qtpy.QtWidgets import QApplication, QMessageBox, QWidget
from enum import Enum


class MessageBoxMode(Enum):
    ERROR = "error"
    WARNING = "warn"
    INFO = "info"
    QUESTION = "question"
    ABOUT = "about"


_QMESSAGE_MODES = {
    MessageBoxMode.ERROR: QMessageBox.critical,
    MessageBoxMode.WARNING: QMessageBox.warning,
    MessageBoxMode.INFO: QMessageBox.information,
    MessageBoxMode.QUESTION: QMessageBox.question,
    MessageBoxMode.ABOUT: QMessageBox.about,
}


def show_messagebox(
    mode: str | MessageBoxMode = MessageBoxMode.INFO,
    title: str = None,
    text: str = None,
    parent=None,
) -> bool:
    """
    Freeze the GUI and open a messagebox dialog.

    Parameters
    ----------
    mode : str or MessageBoxMode, default is MessageBoxMode.INFO
        Mode of message box. Must be "error", "warn", "info", "question" or "about".
    title : str, optional
        Title of messagebox.
    text : str, optional
        Text in messagebox.
    parent : QWidget, optional
        Parent widget.

    Returns
    -------
    bool
        If "OK" or "Yes" is clicked, return True. Otherwise return False.
    """
    show_dialog = _QMESSAGE_MODES[MessageBoxMode(mode)]
    result = show_dialog(parent, title, text)
    return result in (QMessageBox.Ok, QMessageBox.Yes)


def open_url(link: str) -> None:
    """
    Open the link with the default browser.

    Parameters
    ----------
    link : str
        Link to the home page.
    """
    from qtpy.QtGui import QDesktopServices
    from qtpy.QtCore import QUrl

    QDesktopServices.openUrl(QUrl(link))


def screen_center():
    """Get the center coordinate of the screen."""
    return QApplication.desktop().screen().rect().center()


def move_to_screen_center(qwidget: QWidget) -> None:
    """Move a QWidget to the center of screen."""
    qwidget.move(screen_center() - qwidget.rect().center())
    return None


def screen_scale() -> float:
    """Get the scale of main screen."""
    from qtpy.QtGui import QGuiApplication

    screen = QGuiApplication.screens()[0]
    return screen.devicePixelRatio()


def to_clipboard(obj) -> None:
    """
    Copy an object of any type to the clipboard.
    You can copy text, ndarray as an image or data frame as a table data.

    Parameters
    ----------
    obj : Any
        Object to be copied.
    """
    from qtpy.QtGui import QGuiApplication, QImage, qRgb
    import numpy as np
    import pandas as pd

    clipboard = QGuiApplication.clipboard()

    if isinstance(obj, str):
        clipboard.setText(obj)
    elif isinstance(obj, np.ndarray):
        if obj.dtype != np.uint8:
            raise ValueError(f"Cannot copy an array of dtype {obj.dtype} to clipboard.")
        # See https://gist.github.com/smex/5287589
        qimg_format = QImage.Format_RGB888 if obj.ndim == 3 else QImage.Format_Indexed8
        *_, h, w = obj.shape
        qimg = QImage(obj.data, w, h, obj.strides[0], qimg_format)
        gray_color_table = [qRgb(i, i, i) for i in range(256)]
        qimg.setColorTable(gray_color_table)
        clipboard.setImage(qimg)
    elif isinstance(obj, pd.DataFrame):
        clipboard.setText(obj.to_csv(sep="\t"))
    else:
        clipboard.setText(str(obj))
